import math
import warnings

import numpy as np
from numpy import array
import pytest
from pytest import raises as assert_raises

from scipy.fft import fft
from scipy.signal import windows, get_window, resample
from scipy.signal.windows._windows import _WIN_FUNC_DATA, _WIN_FUNCS
from scipy._lib._array_api import (
    xp_assert_close, xp_assert_equal, array_namespace, is_torch, is_jax, is_cupy,
    assert_array_almost_equal, SCIPY_DEVICE, is_numpy, make_xp_test_case,
    make_xp_pytest_param, _xp_copy_to_numpy
)

skip_xp_backends = pytest.mark.skip_xp_backends
xfail_xp_backends = pytest.mark.xfail_xp_backends

lazy_xp_modules = [windows]


window_funcs = [
    ('boxcar', ()),
    ('triang', ()),
    ('parzen', ()),
    ('bohman', ()),
    ('blackman', ()),
    ('nuttall', ()),
    ('blackmanharris', ()),
    ('flattop', ()),
    ('bartlett', ()),
    ('barthann', ()),
    ('hamming', ()),
    ('kaiser', (1,)),
    ('dpss', (2,)),
    ('gaussian', (0.5,)),
    ('general_gaussian', (1.5, 2)),
    ('chebwin', (1,)),
    ('cosine', ()),
    ('hann', ()),
    ('exponential', ()),
    ('taylor', ()),
    ('tukey', (0.5,)),
    ('lanczos', ()),
    ]


@make_xp_test_case(windows.barthann)
class TestBartHann:

    def test_basic(self, xp):
        xp_assert_close(windows.barthann(6, sym=True, xp=xp),
                        xp.asarray([0, 0.35857354213752, 0.8794264578624801,
                         0.8794264578624801, 0.3585735421375199, 0], dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)
        xp_assert_close(windows.barthann(7, xp=xp),
                        xp.asarray([0, 0.27, 0.73, 1.0, 0.73, 0.27, 0],
                                   dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)
        xp_assert_close(windows.barthann(6, False, xp=xp),
                        xp.asarray([0, 0.27, 0.73, 1.0, 0.73, 0.27], dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)


@make_xp_test_case(windows.bartlett)
class TestBartlett:

    def test_basic(self, xp):
        xp_assert_close(windows.bartlett(6, xp=xp),
                        xp.asarray([0, 0.4, 0.8, 0.8, 0.4, 0], dtype=xp.float64))
        xp_assert_close(windows.bartlett(7, xp=xp),
                        xp.asarray([0, 1/3, 2/3, 1.0, 2/3, 1/3, 0], dtype=xp.float64))
        xp_assert_close(windows.bartlett(6, False, xp=xp),
                        xp.asarray([0, 1/3, 2/3, 1.0, 2/3, 1/3], dtype=xp.float64))


@make_xp_test_case(windows.blackman)
class TestBlackman:

    def test_basic(self, xp):
        xp_assert_close(windows.blackman(6, sym=False, xp=xp),
                        xp.asarray([0, 0.13, 0.63, 1.0, 0.63, 0.13], dtype=xp.float64),
                        atol=1e-14)
        xp_assert_close(windows.blackman(7, sym=False, xp=xp),
                        xp.asarray([0, 0.09045342435412804, 0.4591829575459636,
                                    0.9203636180999081, 0.9203636180999081,
                                    0.4591829575459636, 0.09045342435412804],
                                    dtype=xp.float64),
                        atol=1e-8)
        xp_assert_close(windows.blackman(6, xp=xp),
                        xp.asarray([0, 0.2007701432625305, 0.8492298567374694,
                                    0.8492298567374694, 0.2007701432625305, 0],
                                    dtype=xp.float64),
                        atol=1e-14)
        xp_assert_close(windows.blackman(7, True, xp=xp),
                        xp.asarray([0, 0.13, 0.63, 1.0, 0.63, 0.13, 0],
                        dtype=xp.float64), atol=1e-14)


@make_xp_test_case(windows.blackmanharris)
class TestBlackmanHarris:

    def test_basic(self, xp):
        xp_assert_close(windows.blackmanharris(6, False, xp=xp),
                        xp.asarray([6.0e-05, 0.055645, 0.520575,
                                    1.0, 0.520575, 0.055645], dtype=xp.float64))
        xp_assert_close(windows.blackmanharris(7, sym=False, xp=xp),
                        xp.asarray([6.0e-05, 0.03339172347815117, 0.332833504298565,
                                    0.8893697722232837, 0.8893697722232838,
                                    0.3328335042985652, 0.03339172347815122],
                                    dtype=xp.float64))
        xp_assert_close(windows.blackmanharris(6, xp=xp),
                        xp.asarray([6.0e-05, 0.1030114893456638, 0.7938335106543362,
                                    0.7938335106543364, 0.1030114893456638, 6.0e-05],
                                    dtype=xp.float64))
        xp_assert_close(windows.blackmanharris(7, sym=True, xp=xp),
                        xp.asarray([6.0e-05, 0.055645, 0.520575, 1.0, 0.520575,
                                    0.055645, 6.0e-05], dtype=xp.float64))


@make_xp_test_case(windows.taylor)
class TestTaylor:

    def test_normalized(self, xp):
        """Tests windows of small length that are normalized to 1. See the
        documentation for the Taylor window for more information on
        normalization.
        """
        xp_assert_close(windows.taylor(1, 2, 15, xp=xp),
                        xp.asarray([1.0], dtype=xp.float64))
        xp_assert_close(
            windows.taylor(5, 2, 15, xp=xp),
            xp.asarray([0.75803341, 0.90757699, 1.0, 0.90757699, 0.75803341],
                       dtype=xp.float64)
        )
        xp_assert_close(
            windows.taylor(6, 2, 15, xp=xp),
            xp.asarray([
                0.7504082, 0.86624416, 0.98208011, 0.98208011, 0.86624416,
                0.7504082
            ], dtype=xp.float64)
        )

    def test_non_normalized(self, xp):
        """Test windows of small length that are not normalized to 1. See
        the documentation for the Taylor window for more information on
        normalization.
        """
        xp_assert_close(
            windows.taylor(5, 2, 15, norm=False, xp=xp),
            xp.asarray([
                0.87508054, 1.04771499, 1.15440894, 1.04771499, 0.87508054
            ], dtype=xp.float64)
        )
        xp_assert_close(
            windows.taylor(6, 2, 15, norm=False, xp=xp),
            xp.asarray([
                0.86627793, 1.0, 1.13372207, 1.13372207, 1.0, 0.86627793
            ], dtype=xp.float64)
        )

    def test_correctness(self, xp):
        """This test ensures the correctness of the implemented Taylor
        Windowing function. A Taylor Window of 1024 points is created, its FFT
        is taken, and the Peak Sidelobe Level (PSLL) and 3dB and 18dB bandwidth
        are found and checked.

        A publication from Sandia National Laboratories was used as reference
        for the correctness values [1]_.

        References
        -----
        .. [1] Armin Doerry, "Catalog of Window Taper Functions for
               Sidelobe Control", 2017.
               https://www.researchgate.net/profile/Armin_Doerry/publication/316281181_Catalog_of_Window_Taper_Functions_for_Sidelobe_Control/links/58f92cb2a6fdccb121c9d54d/Catalog-of-Window-Taper-Functions-for-Sidelobe-Control.pdf
        """
        M_win = 1024
        N_fft = 131072
        # Set norm=False for correctness as the values obtained from the
        # scientific publication do not normalize the values. Normalizing
        # changes the sidelobe level from the desired value.
        w = windows.taylor(M_win, nbar=4, sll=35, norm=False, sym=False, xp=xp)
        f_np = fft(_xp_copy_to_numpy(w), N_fft)

        spec = 20 * np.log10(np.abs(f_np / np.max(f_np)))

        first_zero = np.argmax(np.diff(spec) > 0)

        PSLL = np.max(spec[first_zero:-first_zero])

        BW_3dB = 2*np.argmax(spec <= -3.0102999566398121) / N_fft * M_win
        BW_18dB = 2*np.argmax(spec <= -18.061799739838872) / N_fft * M_win

        assert math.isclose(PSLL, -35.1672, abs_tol=1)
        assert math.isclose(BW_3dB, 1.1822, abs_tol=0.1)
        assert math.isclose(BW_18dB, 2.6112, abs_tol=0.1)


@make_xp_test_case(windows.bohman)
class TestBohman:

    def test_basic(self, xp):
        xp_assert_close(windows.bohman(6, xp=xp),
                        xp.asarray([0, 0.1791238937062839, 0.8343114522576858,
                                    0.8343114522576858, 0.1791238937062838, 0],
                                    dtype=xp.float64))
        xp_assert_close(windows.bohman(7, sym=True, xp=xp),
                        xp.asarray([0, 0.1089977810442293, 0.6089977810442293, 1.0,
                                    0.6089977810442295, 0.1089977810442293, 0],
                                    dtype=xp.float64))
        xp_assert_close(windows.bohman(6, False, xp=xp),
                        xp.asarray([0, 0.1089977810442293, 0.6089977810442293, 1.0,
                                    0.6089977810442295, 0.1089977810442293],
                                    dtype=xp.float64))


@make_xp_test_case(windows.boxcar)
class TestBoxcar:

    def test_basic(self, xp):
        xp_assert_close(windows.boxcar(6, xp=xp),
                        xp.asarray([1.0, 1, 1, 1, 1, 1], dtype=xp.float64))
        xp_assert_close(windows.boxcar(7, xp=xp),
                        xp.asarray([1.0, 1, 1, 1, 1, 1, 1], dtype=xp.float64))
        xp_assert_close(windows.boxcar(6, False, xp=xp),
                        xp.asarray([1.0, 1, 1, 1, 1, 1], dtype=xp.float64))


cheb_odd_true = [0.200938, 0.107729, 0.134941, 0.165348,
                 0.198891, 0.235450, 0.274846, 0.316836,
                 0.361119, 0.407338, 0.455079, 0.503883,
                 0.553248, 0.602637, 0.651489, 0.699227,
                 0.745266, 0.789028, 0.829947, 0.867485,
                 0.901138, 0.930448, 0.955010, 0.974482,
                 0.988591, 0.997138, 1.000000, 0.997138,
                 0.988591, 0.974482, 0.955010, 0.930448,
                 0.901138, 0.867485, 0.829947, 0.789028,
                 0.745266, 0.699227, 0.651489, 0.602637,
                 0.553248, 0.503883, 0.455079, 0.407338,
                 0.361119, 0.316836, 0.274846, 0.235450,
                 0.198891, 0.165348, 0.134941, 0.107729,
                 0.200938]

cheb_even_true = [0.203894, 0.107279, 0.133904,
                  0.163608, 0.196338, 0.231986,
                  0.270385, 0.311313, 0.354493,
                  0.399594, 0.446233, 0.493983,
                  0.542378, 0.590916, 0.639071,
                  0.686302, 0.732055, 0.775783,
                  0.816944, 0.855021, 0.889525,
                  0.920006, 0.946060, 0.967339,
                  0.983557, 0.994494, 1.000000,
                  1.000000, 0.994494, 0.983557,
                  0.967339, 0.946060, 0.920006,
                  0.889525, 0.855021, 0.816944,
                  0.775783, 0.732055, 0.686302,
                  0.639071, 0.590916, 0.542378,
                  0.493983, 0.446233, 0.399594,
                  0.354493, 0.311313, 0.270385,
                  0.231986, 0.196338, 0.163608,
                  0.133904, 0.107279, 0.203894]


@make_xp_test_case(windows.chebwin)
class TestChebWin:

    def test_basic(self, xp):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            xp_assert_close(windows.chebwin(6, 100, xp=xp),
                            xp.asarray([0.1046401879356917, 0.5075781475823447,
                                        1.0, 1.0,
                                        0.5075781475823447, 0.1046401879356917],
                                        dtype=xp.float64),
                            atol=1e-8
            )
            xp_assert_close(windows.chebwin(7, 100, xp=xp),
                            xp.asarray([0.05650405062850233, 0.316608530648474,
                                        0.7601208123539079, 1.0, 0.7601208123539079,
                                        0.316608530648474, 0.05650405062850233],
                                        dtype=xp.float64))
            xp_assert_close(windows.chebwin(6, 10, xp=xp),
                            xp.asarray([1.0, 0.6071201674458373, 0.6808391469897297,
                                        0.6808391469897297, 0.6071201674458373, 1.0],
                                        dtype=xp.float64))
            xp_assert_close(windows.chebwin(7, 10, xp=xp),
                            xp.asarray([1.0, 0.5190521247588651, 0.5864059018130382,
                                        0.6101519801307441, 0.5864059018130382,
                                        0.5190521247588651, 1.0], dtype=xp.float64))
            xp_assert_close(windows.chebwin(6, 10, False, xp=xp),
                            xp.asarray([1.0, 0.5190521247588651, 0.5864059018130382,
                                        0.6101519801307441, 0.5864059018130382,
                                        0.5190521247588651], dtype=xp.float64))

    def test_cheb_odd_high_attenuation(self, xp):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            cheb_odd = windows.chebwin(53, at=-40, xp=xp)
        assert_array_almost_equal(cheb_odd, xp.asarray(cheb_odd_true), decimal=4)

    def test_cheb_even_high_attenuation(self, xp):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            cheb_even = windows.chebwin(54, at=40, xp=xp)
        assert_array_almost_equal(cheb_even, xp.asarray(cheb_even_true), decimal=4)

    def test_cheb_odd_low_attenuation(self, xp):
        cheb_odd_low_at_true = xp.asarray([1.000000, 0.519052, 0.586405,
                                           0.610151, 0.586405, 0.519052,
                                           1.000000], dtype=xp.float64)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            cheb_odd = windows.chebwin(7, at=10, xp=xp)
        assert_array_almost_equal(cheb_odd, cheb_odd_low_at_true, decimal=4)

    def test_cheb_even_low_attenuation(self, xp):
        cheb_even_low_at_true = xp.asarray([1.000000, 0.451924, 0.51027,
                                            0.541338, 0.541338, 0.51027,
                                            0.451924, 1.000000], dtype=xp.float64)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            cheb_even = windows.chebwin(8, at=-10, xp=xp)
        assert_array_almost_equal(cheb_even, cheb_even_low_at_true, decimal=4)


exponential_data = {
    (4, None, 0.2, False):
        array([4.53999297624848542e-05,
               6.73794699908546700e-03, 1.00000000000000000e+00,
               6.73794699908546700e-03]),
    (4, None, 0.2, True): array([0.00055308437014783, 0.0820849986238988,
                                 0.0820849986238988, 0.00055308437014783]),
    (4, None, 1.0, False): array([0.1353352832366127, 0.36787944117144233, 1.,
                                  0.36787944117144233]),
    (4, None, 1.0, True): array([0.22313016014842982, 0.60653065971263342,
                                 0.60653065971263342, 0.22313016014842982]),
    (4, 2, 0.2, False):
        array([4.53999297624848542e-05, 6.73794699908546700e-03,
               1.00000000000000000e+00, 6.73794699908546700e-03]),
    (4, 2, 0.2, True): None,
    (4, 2, 1.0, False): array([0.1353352832366127, 0.36787944117144233, 1.,
                               0.36787944117144233]),
    (4, 2, 1.0, True): None,
    (5, None, 0.2, True):
        array([4.53999297624848542e-05,
               6.73794699908546700e-03, 1.00000000000000000e+00,
               6.73794699908546700e-03, 4.53999297624848542e-05]),
    (5, None, 1.0, True): array([0.1353352832366127, 0.36787944117144233, 1.,
                                 0.36787944117144233, 0.1353352832366127]),
    (5, 2, 0.2, True): None,
    (5, 2, 1.0, True): None
}


@make_xp_test_case(windows.exponential)
def test_exponential(xp):
    for k, v in exponential_data.items():
        if v is None:
            assert_raises(ValueError, windows.exponential, *k, xp=xp)
        else:
            win = windows.exponential(*k, xp=xp)
            xp_assert_close(win, xp.asarray(v), rtol=1e-14)


@make_xp_test_case(windows.flattop)
class TestFlatTop:

    def test_basic(self, xp):
        xp_assert_close(windows.flattop(6, sym=False, xp=xp),
                        xp.asarray([-0.000421051, -0.051263156, 0.19821053, 1.0,
                                     0.19821053, -0.051263156], dtype=xp.float64))
        xp_assert_close(windows.flattop(7, sym=False, xp=xp),
                        xp.asarray([-0.000421051, -0.03684078115492348,
                                     0.01070371671615342, 0.7808739149387698,
                                     0.7808739149387698, 0.01070371671615342,
                                    -0.03684078115492348], dtype=xp.float64))
        xp_assert_close(windows.flattop(6, xp=xp),
                        xp.asarray([-0.000421051, -0.0677142520762119,
                                     0.6068721525762117, 0.6068721525762117,
                                    -0.0677142520762119, -0.000421051],
                                    dtype=xp.float64))
        xp_assert_close(windows.flattop(7, True, xp=xp),
                        xp.asarray([-0.000421051, -0.051263156, 0.19821053, 1.0,
                                     0.19821053, -0.051263156, -0.000421051],
                                     dtype=xp.float64))


@make_xp_test_case(windows.gaussian)
class TestGaussian:

    def test_basic(self, xp):
        xp_assert_close(windows.gaussian(6, 1.0, xp=xp),
                        xp.asarray([0.04393693362340742, 0.3246524673583497,
                                    0.8824969025845955, 0.8824969025845955,
                                    0.3246524673583497, 0.04393693362340742],
                                    dtype=xp.float64))
        xp_assert_close(windows.gaussian(7, 1.2, xp=xp),
                        xp.asarray([0.04393693362340742, 0.2493522087772962,
                                    0.7066482778577162, 1.0, 0.7066482778577162,
                                    0.2493522087772962, 0.04393693362340742],
                                    dtype=xp.float64))
        xp_assert_close(windows.gaussian(7, 3, xp=xp),
                        xp.asarray([0.6065306597126334, 0.8007374029168081,
                                    0.9459594689067654, 1.0, 0.9459594689067654,
                                    0.8007374029168081, 0.6065306597126334],
                                    dtype=xp.float64))
        xp_assert_close(windows.gaussian(6, 3, False, xp=xp),
                        xp.asarray([0.6065306597126334, 0.8007374029168081,
                                    0.9459594689067654, 1.0, 0.9459594689067654,
                                    0.8007374029168081], dtype=xp.float64))


@make_xp_test_case(windows.general_cosine)
class TestGeneralCosine:

    def test_basic(self, xp):
        a = xp.asarray([0.5, 0.3, 0.2])
        xp_assert_close(windows.general_cosine(5, a),
                        xp.asarray([0.4, 0.3, 1, 0.3, 0.4], dtype=xp.float64))

        a = xp.asarray([0.5, 0.3, 0.2])
        xp_assert_close(windows.general_cosine(4, a, sym=False),
                        xp.asarray([0.4, 0.3, 1, 0.3], dtype=xp.float64))


@make_xp_test_case(windows.general_hamming)
class TestGeneralHamming:

    def test_basic(self, xp):
        xp_assert_close(windows.general_hamming(5, 0.7, xp=xp),
                        xp.asarray([0.4, 0.7, 1.0, 0.7, 0.4], dtype=xp.float64))
        xp_assert_close(windows.general_hamming(5, 0.75, sym=False, xp=xp),
                        xp.asarray([0.5, 0.6727457514, 0.9522542486,
                                    0.9522542486, 0.6727457514], dtype=xp.float64))
        xp_assert_close(windows.general_hamming(6, 0.75, sym=True, xp=xp),
                        xp.asarray([0.5, 0.6727457514, 0.9522542486,
                                    0.9522542486, 0.6727457514, 0.5], dtype=xp.float64))


@make_xp_test_case(windows.hamming)
class TestHamming:

    def test_basic(self, xp):
        xp_assert_close(windows.hamming(6, False, xp=xp),
                        xp.asarray([0.08, 0.31, 0.77, 1.0, 0.77, 0.31],
                                   dtype=xp.float64))
        xp_assert_close(windows.hamming(7, sym=False, xp=xp),
                        xp.asarray([0.08, 0.2531946911449826, 0.6423596296199047,
                                    0.9544456792351128, 0.9544456792351128,
                                    0.6423596296199047, 0.2531946911449826],
                                    dtype=xp.float64))
        xp_assert_close(windows.hamming(6, xp=xp),
                        xp.asarray([0.08, 0.3978521825875242, 0.9121478174124757,
                                    0.9121478174124757, 0.3978521825875242, 0.08],
                                    dtype=xp.float64))
        xp_assert_close(windows.hamming(7, sym=True, xp=xp),
                        xp.asarray([0.08, 0.31, 0.77, 1.0, 0.77, 0.31, 0.08],
                                   dtype=xp.float64))


@make_xp_test_case(windows.hann)
class TestHann:

    def test_basic(self, xp):
        xp_assert_close(windows.hann(6, sym=False, xp=xp),
                        xp.asarray([0, 0.25, 0.75, 1.0, 0.75, 0.25], dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)
        xp_assert_close(windows.hann(7, sym=False, xp=xp),
                        xp.asarray([0, 0.1882550990706332, 0.6112604669781572,
                                    0.9504844339512095, 0.9504844339512095,
                                    0.6112604669781572, 0.1882550990706332],
                                    dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)
        xp_assert_close(windows.hann(6, True, xp=xp),
                        xp.asarray([0, 0.3454915028125263, 0.9045084971874737,
                                    0.9045084971874737, 0.3454915028125263, 0],
                                    dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)
        xp_assert_close(windows.hann(7, xp=xp),
                        xp.asarray([0, 0.25, 0.75, 1.0, 0.75, 0.25, 0],
                        dtype=xp.float64),
                        rtol=1e-15, atol=1e-15)


@make_xp_test_case(windows.kaiser)
class TestKaiser:

    def test_basic(self, xp):
        xp_assert_close(windows.kaiser(6, 0.5, xp=xp),
                        xp.asarray([0.9403061933191572, 0.9782962393705389,
                                    0.9975765035372042, 0.9975765035372042,
                                    0.9782962393705389, 0.9403061933191572],
                                    dtype=xp.float64))
        xp_assert_close(windows.kaiser(7, 0.5, xp=xp),
                        xp.asarray([0.9403061933191572, 0.9732402256999829,
                                    0.9932754654413773, 1.0, 0.9932754654413773,
                                    0.9732402256999829, 0.9403061933191572],
                                    dtype=xp.float64))
        xp_assert_close(windows.kaiser(6, 2.7, xp=xp),
                        xp.asarray([0.2603047507678832, 0.6648106293528054,
                                    0.9582099802511439, 0.9582099802511439,
                                    0.6648106293528054, 0.2603047507678832],
                                    dtype=xp.float64))
        xp_assert_close(windows.kaiser(7, 2.7, xp=xp),
                        xp.asarray([0.2603047507678832, 0.5985765418119844,
                                    0.8868495172060835, 1.0, 0.8868495172060835,
                                    0.5985765418119844, 0.2603047507678832],
                                    dtype=xp.float64))
        xp_assert_close(windows.kaiser(6, 2.7, False, xp=xp),
                        xp.asarray([0.2603047507678832, 0.5985765418119844,
                                    0.8868495172060835, 1.0, 0.8868495172060835,
                                    0.5985765418119844], dtype=xp.float64))


@make_xp_test_case(windows.kaiser_bessel_derived)
class TestKaiserBesselDerived:

    def test_basic(self, xp):
        # cover case `M < 1`
        w = windows.kaiser_bessel_derived(0.5, beta=4.0, xp=xp)
        xp_assert_equal(w, xp.asarray([]))

        M = 100
        w = windows.kaiser_bessel_derived(M, beta=4.0, xp=xp)
        w2 = windows.get_window(('kaiser bessel derived', 4.0),
                                M, fftbins=False, xp=xp)
        xp_assert_close(w, w2)

        # Test for Princen-Bradley condition
        actual = w[:M // 2] ** 2 + w[-M // 2:] ** 2
        xp_assert_close(actual, xp.ones(actual.shape, dtype=actual.dtype))

        # Test actual values from other implementations
        # M = 2:  sqrt(2) / 2
        # M = 4:  0.518562710536, 0.855039598640
        # M = 6:  0.436168993154, 0.707106781187, 0.899864772847
        # Ref:https://github.com/scipy/scipy/pull/4747#issuecomment-172849418
        actual = windows.kaiser_bessel_derived(2, beta=np.pi / 2, xp=xp)[:1]
        desired = xp.ones_like(actual) * math.sqrt(2) / 2.0
        xp_assert_close(actual, desired)

        xp_assert_close(windows.kaiser_bessel_derived(4, beta=np.pi / 2, xp=xp)[:2],
                        xp.asarray([0.518562710536, 0.855039598640], dtype=xp.float64))

        xp_assert_close(windows.kaiser_bessel_derived(6, beta=np.pi / 2, xp=xp)[:3],
                        xp.asarray([0.436168993154, 0.707106781187, 0.899864772847],
                                   dtype=xp.float64))

    def test_exceptions(self, xp):
        M = 100
        # Assert ValueError for odd window length
        msg = ("Kaiser-Bessel Derived windows are only defined for even "
               "number of points")
        with assert_raises(ValueError, match=msg):
            windows.kaiser_bessel_derived(M + 1, beta=4., xp=xp)

        # Assert ValueError for non-symmetric setting
        msg = ("Kaiser-Bessel Derived windows are only defined for "
               "symmetric shapes")
        with assert_raises(ValueError, match=msg):
            windows.kaiser_bessel_derived(M + 1, beta=4., sym=False, xp=xp)


@make_xp_test_case(windows.nuttall)
class TestNuttall:

    def test_basic(self, xp):
        xp_assert_close(windows.nuttall(6, sym=False, xp=xp),
                        xp.asarray([0.0003628, 0.0613345, 0.5292298, 1.0, 0.5292298,
                                    0.0613345], dtype=xp.float64))
        xp_assert_close(windows.nuttall(7, sym=False, xp=xp),
                        xp.asarray([0.0003628, 0.03777576895352025,
                                    0.3427276199688195,
                                    0.8918518610776603, 0.8918518610776603,
                                    0.3427276199688196, 0.0377757689535203],
                                    dtype=xp.float64))
        xp_assert_close(windows.nuttall(6, xp=xp),
                        xp.asarray([0.0003628, 0.1105152530498718,
                                    0.7982580969501282, 0.7982580969501283,
                                    0.1105152530498719, 0.0003628], dtype=xp.float64))
        xp_assert_close(windows.nuttall(7, True, xp=xp),
                        xp.asarray([0.0003628, 0.0613345, 0.5292298, 1.0,
                                    0.5292298, 0.0613345, 0.0003628], dtype=xp.float64))


@make_xp_test_case(windows.parzen)
class TestParzen:

    def test_basic(self, xp):
        xp_assert_close(windows.parzen(6, xp=xp),
                        xp.asarray([0.009259259259259254, 0.25, 0.8611111111111112,
                                    0.8611111111111112, 0.25, 0.009259259259259254],
                                    dtype=xp.float64))
        xp_assert_close(windows.parzen(7, sym=True, xp=xp),
                        xp.asarray([0.00583090379008747, 0.1574344023323616,
                                    0.6501457725947521, 1.0, 0.6501457725947521,
                                    0.1574344023323616, 0.00583090379008747],
                                    dtype=xp.float64))
        xp_assert_close(windows.parzen(6, False, xp=xp),
                        xp.asarray([0.00583090379008747, 0.1574344023323616,
                                    0.6501457725947521, 1.0, 0.6501457725947521,
                                    0.1574344023323616], dtype=xp.float64))


@make_xp_test_case(windows.triang)
class TestTriang:

    def test_basic(self, xp):

        xp_assert_close(windows.triang(6, True, xp=xp),
                        xp.asarray([1/6, 1/2, 5/6, 5/6, 1/2, 1/6], dtype=xp.float64))
        xp_assert_close(windows.triang(7, xp=xp),
                        xp.asarray([1/4, 1/2, 3/4, 1, 3/4, 1/2, 1/4], dtype=xp.float64))
        xp_assert_close(windows.triang(6, sym=False, xp=xp),
                        xp.asarray([1/4, 1/2, 3/4, 1, 3/4, 1/2], dtype=xp.float64))


tukey_data = {
    (4, 0.5, True): array([0.0, 1.0, 1.0, 0.0]),
    (4, 0.9, True): array([0.0, 0.84312081893436686,
                           0.84312081893436686, 0.0]),
    (4, 1.0, True): array([0.0, 0.75, 0.75, 0.0]),
    (4, 0.5, False): array([0.0, 1.0, 1.0, 1.0]),
    (4, 0.9, False): array([0.0, 0.58682408883346526,
                            1.0, 0.58682408883346526]),
    (4, 1.0, False): array([0.0, 0.5, 1.0, 0.5]),
    (5, 0.0, True): array([1.0, 1.0, 1.0, 1.0, 1.0]),
    (5, 0.8, True): array([0.0, 0.69134171618254492,
                           1.0, 0.69134171618254492, 0.0]),
    (5, 1.0, True): array([0.0, 0.5, 1.0, 0.5, 0.0]),

    (6, 0): [1.0, 1, 1, 1, 1, 1],
    (7, 0): [1.0, 1, 1, 1, 1, 1, 1],
    (6, .25): [0.0, 1, 1, 1, 1, 0],
    (7, .25): [0.0, 1, 1, 1, 1, 1, 0],
    (6,): [0, 0.9045084971874737, 1.0, 1.0, 0.9045084971874735, 0],
    (7,): [0, 0.75, 1.0, 1.0, 1.0, 0.75, 0],
    (6, .75): [0, 0.5522642316338269, 1.0, 1.0, 0.5522642316338267, 0],
    (7, .75): [0, 0.4131759111665348, 0.9698463103929542, 1.0,
               0.9698463103929542, 0.4131759111665347, 0],
    (6, 1): [0, 0.3454915028125263, 0.9045084971874737, 0.9045084971874737,
             0.3454915028125263, 0],
    (7, 1): [0, 0.25, 0.75, 1.0, 0.75, 0.25, 0],
}


@make_xp_test_case(windows.tukey)
class TestTukey:

    def test_basic(self, xp):
        # Test against hardcoded data
        for k, v in tukey_data.items():
            if v is None:
                assert_raises(ValueError, windows.tukey, *k, xp=xp)
            else:
                if is_torch(xp) and k in [(6,), (6, .75), (7, .75), (6,1)]:
                     atol_rtol = {'rtol': 3e-8, 'atol': 1e-8}
                else:
                     atol_rtol = {'rtol': 1e-15, 'atol': 1e-15 }

                win = windows.tukey(*k, xp=xp)
                xp_assert_close(win, xp.asarray(v),
                                check_dtype=False, **atol_rtol)

    def test_extremes(self, xp):
        # Test extremes of alpha correspond to boxcar and hann
        tuk0 = windows.tukey(100, 0, xp=xp)
        box0 = windows.boxcar(100, xp=xp)
        xp_assert_close(tuk0, box0)

        tuk1 = windows.tukey(100, 1, xp=xp)
        han1 = windows.hann(100, xp=xp)
        xp_assert_close(tuk1, han1)


dpss_data = {
    # All values from MATLAB:
    # * taper[1] of (3, 1.4, 3) sign-flipped
    # * taper[3] of (5, 1.5, 5) sign-flipped
    (4, 0.1, 2): ([[0.497943898, 0.502047681, 0.502047681, 0.497943898], [0.670487993, 0.224601537, -0.224601537, -0.670487993]], [0.197961815, 0.002035474]),  # noqa: E501
    (3, 1.4, 3): ([[0.410233151, 0.814504464, 0.410233151], [0.707106781, 0.0, -0.707106781], [0.575941629, -0.580157287, 0.575941629]], [0.999998093, 0.998067480, 0.801934426]),  # noqa: E501
    (5, 1.5, 5): ([[0.1745071052, 0.4956749177, 0.669109327, 0.495674917, 0.174507105], [0.4399493348, 0.553574369, 0.0, -0.553574369, -0.439949334], [0.631452756, 0.073280238, -0.437943884, 0.073280238, 0.631452756], [0.553574369, -0.439949334, 0.0, 0.439949334, -0.553574369], [0.266110290, -0.498935248, 0.600414741, -0.498935248, 0.266110290147157]], [0.999728571, 0.983706916, 0.768457889, 0.234159338, 0.013947282907567]),  # noqa: E501
    (100, 2, 4): ([[0.0030914414, 0.0041266922, 0.005315076, 0.006665149, 0.008184854, 0.0098814158, 0.011761239, 0.013829809, 0.016091597, 0.018549973, 0.02120712, 0.02406396, 0.027120092, 0.030373728, 0.033821651, 0.037459181, 0.041280145, 0.045276872, 0.049440192, 0.053759447, 0.058222524, 0.062815894, 0.067524661, 0.072332638, 0.077222418, 0.082175473, 0.087172252, 0.092192299, 0.097214376, 0.1022166, 0.10717657, 0.11207154, 0.11687856, 0.12157463, 0.12613686, 0.13054266, 0.13476986, 0.13879691, 0.14260302, 0.14616832, 0.14947401, 0.1525025, 0.15523755, 0.15766438, 0.15976981, 0.16154233, 0.16297223, 0.16405162, 0.16477455, 0.16513702, 0.16513702, 0.16477455, 0.16405162, 0.16297223, 0.16154233, 0.15976981, 0.15766438, 0.15523755, 0.1525025, 0.14947401, 0.14616832, 0.14260302, 0.13879691, 0.13476986, 0.13054266, 0.12613686, 0.12157463, 0.11687856, 0.11207154, 0.10717657, 0.1022166, 0.097214376, 0.092192299, 0.087172252, 0.082175473, 0.077222418, 0.072332638, 0.067524661, 0.062815894, 0.058222524, 0.053759447, 0.049440192, 0.045276872, 0.041280145, 0.037459181, 0.033821651, 0.030373728, 0.027120092, 0.02406396, 0.02120712, 0.018549973, 0.016091597, 0.013829809, 0.011761239, 0.0098814158, 0.008184854, 0.006665149, 0.005315076, 0.0041266922, 0.0030914414], [0.018064449, 0.022040342, 0.026325013, 0.030905288, 0.035764398, 0.040881982, 0.046234148, 0.051793558, 0.057529559, 0.063408356, 0.069393216, 0.075444716, 0.081521022, 0.087578202, 0.093570567, 0.099451049, 0.10517159, 0.11068356, 0.11593818, 0.12088699, 0.12548227, 0.12967752, 0.1334279, 0.13669069, 0.13942569, 0.1415957, 0.14316686, 0.14410905, 0.14439626, 0.14400686, 0.14292389, 0.1411353, 0.13863416, 0.13541876, 0.13149274, 0.12686516, 0.12155045, 0.1155684, 0.10894403, 0.10170748, 0.093893752, 0.08554251, 0.076697768, 0.067407559, 0.057723559, 0.04770068, 0.037396627, 0.026871428, 0.016186944, 0.0054063557, -0.0054063557, -0.016186944, -0.026871428, -0.037396627, -0.04770068, -0.057723559, -0.067407559, -0.076697768, -0.08554251, -0.093893752, -0.10170748, -0.10894403, -0.1155684, -0.12155045, -0.12686516, -0.13149274, -0.13541876, -0.13863416, -0.1411353, -0.14292389, -0.14400686, -0.14439626, -0.14410905, -0.14316686, -0.1415957, -0.13942569, -0.13669069, -0.1334279, -0.12967752, -0.12548227, -0.12088699, -0.11593818, -0.11068356, -0.10517159, -0.099451049, -0.093570567, -0.087578202, -0.081521022, -0.075444716, -0.069393216, -0.063408356, -0.057529559, -0.051793558, -0.046234148, -0.040881982, -0.035764398, -0.030905288, -0.026325013, -0.022040342, -0.018064449], [0.064817553, 0.072567801, 0.080292992, 0.087918235, 0.095367076, 0.10256232, 0.10942687, 0.1158846, 0.12186124, 0.12728523, 0.13208858, 0.13620771, 0.13958427, 0.14216587, 0.14390678, 0.14476863, 0.1447209, 0.14374148, 0.14181704, 0.13894336, 0.13512554, 0.13037812, 0.1247251, 0.11819984, 0.11084487, 0.10271159, 0.093859853, 0.084357497, 0.074279719, 0.063708406, 0.052731374, 0.041441525, 0.029935953, 0.018314987, 0.0066811877, -0.0048616765, -0.016209689, -0.027259848, -0.037911124, -0.048065512, -0.05762905, -0.066512804, -0.0746338, -0.081915903, -0.088290621, -0.09369783, -0.098086416, -0.10141482, -0.10365146, -0.10477512, -0.10477512, -0.10365146, -0.10141482, -0.098086416, -0.09369783, -0.088290621, -0.081915903, -0.0746338, -0.066512804, -0.05762905, -0.048065512, -0.037911124, -0.027259848, -0.016209689, -0.0048616765, 0.0066811877, 0.018314987, 0.029935953, 0.041441525, 0.052731374, 0.063708406, 0.074279719, 0.084357497, 0.093859853, 0.10271159, 0.11084487, 0.11819984, 0.1247251, 0.13037812, 0.13512554, 0.13894336, 0.14181704, 0.14374148, 0.1447209, 0.14476863, 0.14390678, 0.14216587, 0.13958427, 0.13620771, 0.13208858, 0.12728523, 0.12186124, 0.1158846, 0.10942687, 0.10256232, 0.095367076, 0.087918235, 0.080292992, 0.072567801, 0.064817553], [0.14985551, 0.15512305, 0.15931467, 0.16236806, 0.16423291, 0.16487165, 0.16426009, 0.1623879, 0.1592589, 0.15489114, 0.14931693, 0.14258255, 0.13474785, 0.1258857, 0.11608124, 0.10543095, 0.094041635, 0.082029213, 0.069517411, 0.056636348, 0.043521028, 0.030309756, 0.017142511, 0.0041592774, -0.0085016282, -0.020705223, -0.032321494, -0.043226982, -0.053306291, -0.062453515, -0.070573544, -0.077583253, -0.083412547, -0.088005244, -0.091319802, -0.093329861, -0.094024602, -0.093408915, -0.091503383, -0.08834406, -0.08398207, -0.078483012, -0.071926192, -0.064403681, -0.056019215, -0.046886954, -0.037130106, -0.026879442, -0.016271713, -0.005448, 0.005448, 0.016271713, 0.026879442, 0.037130106, 0.046886954, 0.056019215, 0.064403681, 0.071926192, 0.078483012, 0.08398207, 0.08834406, 0.091503383, 0.093408915, 0.094024602, 0.093329861, 0.091319802, 0.088005244, 0.083412547, 0.077583253, 0.070573544, 0.062453515, 0.053306291, 0.043226982, 0.032321494, 0.020705223, 0.0085016282, -0.0041592774, -0.017142511, -0.030309756, -0.043521028, -0.056636348, -0.069517411, -0.082029213, -0.094041635, -0.10543095, -0.11608124, -0.1258857, -0.13474785, -0.14258255, -0.14931693, -0.15489114, -0.1592589, -0.1623879, -0.16426009, -0.16487165, -0.16423291, -0.16236806, -0.15931467, -0.15512305, -0.14985551]], [0.999943140, 0.997571533, 0.959465463, 0.721862496]),  # noqa: E501
}


@make_xp_test_case(windows.dpss)
class TestDPSS:

    def test_basic(self, xp):
        # Test against hardcoded data
        for k, v in dpss_data.items():
            win, ratios = windows.dpss(*k, return_ratios=True, xp=xp)
            xp_assert_close(win, v[0], atol=1e-7, err_msg=k)
            xp_assert_close(ratios, v[1], rtol=1e-5, atol=1e-7, err_msg=k)

    def test_unity(self, xp):
        # Test unity value handling (gh-2221)
        for M in range(1, 21):
            # corrected w/approximation (default)
            win = windows.dpss(M, M / 2.1, xp=xp)
            expected = M % 2  # one for odd, none for even
            xp_assert_equal(np.isclose(win, 1.).sum(), expected,
                         err_msg=f'{win}')
            # corrected w/subsample delay (slower)
            win_sub = windows.dpss(M, M / 2.1, norm='subsample', xp=xp)
            if M > 2:
                # @M=2 the subsample doesn't do anything
                xp_assert_equal(np.isclose(win_sub, 1.).sum(), expected,
                             err_msg=f'{win_sub}')
                xp_assert_close(win, win_sub, rtol=0.03)  # within 3%
            # not the same, l2-norm
            win_2 = windows.dpss(M, M / 2.1, norm=2, xp=xp)
            expected = 1 if M == 1 else 0
            xp_assert_equal(np.isclose(win_2, 1.).sum(), expected,
                         err_msg=f'{win_2}')

    def test_extremes(self, xp):
        # Test extremes of alpha
        lam = windows.dpss(31, 6, 4, return_ratios=True, xp=xp)[1]
        xp_assert_close(lam, xp.ones_like(lam))
        lam = windows.dpss(31, 7, 4, return_ratios=True, xp=xp)[1]
        xp_assert_close(lam, xp.ones_like(lam))
        lam = windows.dpss(31, 8, 4, return_ratios=True, xp=xp)[1]
        xp_assert_close(lam, xp.ones_like(lam))

    def test_degenerate(self, xp):
        # Test failures
        assert_raises(ValueError, windows.dpss, 4, 1.5, -1)  # Bad Kmax
        assert_raises(ValueError, windows.dpss, 4, 1.5, -5)
        assert_raises(TypeError, windows.dpss, 4, 1.5, 1.1)
        assert_raises(ValueError, windows.dpss, 3, 1.5, 3)  # NW must be < N/2.
        assert_raises(ValueError, windows.dpss, 3, -1, 3)  # NW must be pos
        assert_raises(ValueError, windows.dpss, 3, 0, 3)
        assert_raises(ValueError, windows.dpss, -1, 1, 3)  # negative M

    @skip_xp_backends(np_only=True)
    def test_degenerate_signle_samples(self, xp):
        # Single samples
        w = windows.dpss(1, 1.)
        xp_assert_equal(w, [1.])
        w, ratio = windows.dpss(1, 1., return_ratios=True)
        xp_assert_equal(w, [1.])
        assert ratio == 1.
        w, ratio = windows.dpss(1, 1., Kmax=4, return_ratios=True)
        xp_assert_equal(w, [1.])
        assert isinstance(ratio, np.ndarray)
        xp_assert_equal(ratio, [1.])

        assert_raises(ValueError, windows.dpss, 4, 1.5, -1, xp=xp)  # Bad Kmax
        assert_raises(ValueError, windows.dpss, 4, 1.5, -5, xp=xp)
        assert_raises(TypeError, windows.dpss, 4, 1.5, 1.1, xp=xp)
        assert_raises(ValueError, windows.dpss, 3, 1.5, 3, xp=xp)  # NW must be < N/2.
        assert_raises(ValueError, windows.dpss, 3, -1, 3, xp=xp)  # NW must be pos
        assert_raises(ValueError, windows.dpss, 3, 0, 3, xp=xp)
        assert_raises(ValueError, windows.dpss, -1, 1, 3, xp=xp)  # negative M


@make_xp_test_case(windows.lanczos)
class TestLanczos:

    def test_basic(self, xp):
        # Analytical results:
        # sinc(x) = sinc(-x)
        # sinc(pi) = 0, sinc(0) = 1
        # Hand computation on WolframAlpha:
        # sinc(2 pi / 3) = 0.413496672
        # sinc(pi / 3) = 0.826993343
        # sinc(3 pi / 5) = 0.504551152
        # sinc(pi / 5) = 0.935489284
        xp_assert_close(windows.lanczos(6, sym=False, xp=xp),
                        xp.asarray([0., 0.413496672,
                                    0.826993343, 1., 0.826993343,
                                    0.413496672], dtype=xp.float64),
                        atol=1e-9)
        xp_assert_close(windows.lanczos(6, xp=xp),
                        xp.asarray([0., 0.504551152,
                                    0.935489284, 0.935489284,
                                    0.504551152, 0.], dtype=xp.float64),
                        atol=1e-9)
        xp_assert_close(windows.lanczos(7, sym=True, xp=xp),
                        xp.asarray([0., 0.413496672,
                                    0.826993343, 1., 0.826993343,
                                    0.413496672, 0.], dtype=xp.float64),
                        atol=1e-9)

    def test_array_size(self, xp):
        for n in [0, 10, 11]:
            assert windows.lanczos(n, sym=False, xp=xp).shape[0] == n
            assert windows.lanczos(n, sym=True, xp=xp).shape[0] == n


@make_xp_test_case(windows.get_window)
class TestGetWindow:
    """Unit test for `scipy.signal.get_windows`. """

    def test_WIN_FUNC_DATA_integrity(self):
        """Verify that the `_windows._WIN_FUNC_DATA` dict is consistent.

          The keys of _WIN_FUNC_DATA are made of tuples of strings of allowed window
          names. Its values are 2-tuples made up of the window function and a
          entry characterizing the existence of window parameters as ``True``,
          ``False`` or ``'OPTIONAL'``.


          It is verified that the correct window name (i.e., corresponding to the
          function in the value tuple) is included in the key tuple. It is also checked
          that the second entry in the value tuple is either ``True``, ``False`` or
          ``'OPTIONAL'``.
          """
        for nn_, v_ in _WIN_FUNC_DATA.items():
            func_name = v_[0].__name__
            msg = f"Function name in {nn_} does not contain name of actual function!"
            assert func_name in nn_, msg
            assert v_[1] in (True, False, 'OPTIONAL')

    @make_xp_test_case(windows.boxcar)
    def test_boxcar(self, xp):
        w = windows.get_window('boxcar', 12, xp=xp)
        xp_assert_equal(w, xp.ones_like(w))

        # window is a tuple of len 1
        w = windows.get_window(('boxcar',), 16, xp=xp)
        xp_assert_equal(w, xp.ones_like(w))

    @make_xp_test_case(windows.chebwin)
    def test_cheb_odd(self, xp):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            w = windows.get_window(('chebwin', -40), 53, fftbins=False, xp=xp)
        assert_array_almost_equal(
            w, xp.asarray(cheb_odd_true, dtype=xp.float64), decimal=4
        )

    @make_xp_test_case(windows.chebwin)
    def test_cheb_even(self, xp):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "This window is not suitable", UserWarning)
            w = windows.get_window(('chebwin', 40), 54, fftbins=False, xp=xp)
        assert_array_almost_equal(w, xp.asarray(cheb_even_true), decimal=4)

    @make_xp_test_case(windows.dpss)
    def test_dpss(self, xp):
        win1 = windows.get_window(('dpss', 3), 64, fftbins=False, xp=xp)
        win2 = windows.dpss(64, 3, xp=xp)
        xp_assert_equal(win1, win2)

    @make_xp_test_case(windows.kaiser)
    def test_kaiser_float(self, xp):
        win1 = windows.get_window(7.2, 64, xp=xp)
        win2 = windows.kaiser(64, 7.2, False, xp=xp)
        if is_jax(xp):
            # On JAX with jit enabled, there is a very small discrepancy
            # in the results.
            xp_assert_close(win1, win2, rtol=xp.finfo(win1.dtype).eps)
        else:
            xp_assert_equal(win1, win2)

    @pytest.mark.parametrize('Nx', [-1, 3.0, np.float64(3)])
    @make_xp_test_case(windows.hann)
    def test_invalid_parameter_NX(self, Nx, xp):
        with pytest.raises(ValueError, match="^Parameter Nx=.*"):
            windows.get_window('hann', Nx, xp=xp)

    # noinspection PyTypeChecker
    def test_invalid_inputs(self, xp):
        """Raise all exceptions (except those concerning parameter `Nx`). """
        with pytest.raises(ValueError, match="^Parameter fftbins=.*"):
            windows.get_window('hann', 5, fftbins=1, xp=xp)
        with pytest.raises(ValueError, match="^Parameter window=.*"):
            windows.get_window(['hann',], 5, xp=xp)
        with pytest.raises(ValueError, match="^First tuple entry of parameter win.*"):
            windows.get_window((42,), 5, xp=xp)
        with pytest.raises(ValueError, match="^Invalid window name 'INVALID'.*"):
            windows.get_window('INVALID', 5, xp=xp)
        with pytest.raises(ValueError, match="^'hann' does not allow parameters.*"):
            windows.get_window(('hann', 1), 5, xp=xp)
        with pytest.raises(ValueError, match="^'kaiser' must have parameters.*"):
            windows.get_window('kaiser', 5, xp=xp)
        with pytest.raises(ValueError, match="^Window dpss must have one.*"):
            windows.get_window(('dpss', 1, 2), 5, xp=xp)
        with pytest.raises(ValueError, match="^'general_cosine' does not accept.*"):
            xp_ = xp or np  # ensure parameter xp_ is not None
            windows.get_window(('general cosine', [1, 2]), 5, xp=xp_)

    @make_xp_test_case(windows.bartlett)
    def test_symmetric_periodic(self, xp):
        """Ensure that suffixes `_periodic` and `_symmetric` work for window names. """
        w_sym = windows.bartlett(5, sym=True, xp=xp)
        xp_assert_close(get_window('bartlett', 5, fftbins=False, xp=xp), w_sym)
        xp_assert_close(get_window('bartlett_symmetric', 5, xp=xp), w_sym)
        # overwrite parameter `fftbins`:
        xp_assert_close(get_window('bartlett_symmetric', 5, fftbins=True, xp=xp), w_sym)

        w_per = windows.bartlett(5, sym=False, xp=xp)
        xp_assert_close(get_window('bartlett', 5, xp=xp), w_per)
        xp_assert_close(get_window('bartlett', 5, fftbins=True, xp=xp), w_per)
        xp_assert_close(get_window('bartlett_periodic', 5, xp=xp), w_per)
        # overwrite parameter `fftbins`:
        xp_assert_close(get_window('bartlett_periodic', 5, fftbins=False, xp=xp),
                        w_per)

    @make_xp_test_case(windows.kaiser)
    def test_array_as_window(self, xp):
        # github issue 3603
        osfactor = 128
        sig = xp.arange(128)

        win = windows.get_window(('kaiser', 8.0), osfactor // 2, xp=xp)
        mesg = "^window must|^window.shape="
        with assert_raises(ValueError, match=mesg):
            resample(sig, sig.shape[0] * osfactor, window=win)

    @make_xp_test_case(windows.general_cosine)
    def test_general_cosine(self, xp):
        xp_assert_close(get_window(('general_cosine', xp.asarray([0.5, 0.3, 0.2])), 4),
                        xp.asarray([0.4, 0.3, 1, 0.3], dtype=xp.float64))
        xp_assert_close(get_window(('general_cosine', xp.asarray([0.5, 0.3, 0.2])), 4,
                                   fftbins=False),
                        xp.asarray([0.4, 0.55, 0.55, 0.4], dtype=xp.float64))

        with pytest.raises(ValueError):
            get_window(('general_cosine', [0.5, 0.3, 0.2]), 4, xp=xp)

    @make_xp_test_case(windows.general_hamming)
    def test_general_hamming(self, xp):
        xp_assert_close(get_window(('general_hamming', 0.7), 5, xp=xp),
                        xp.asarray([0.4, 0.6072949, 0.9427051, 0.9427051, 0.6072949],
                                   dtype=xp.float64))
        xp_assert_close(get_window(('general_hamming', 0.7), 5, fftbins=False, xp=xp),
                        xp.asarray([0.4, 0.7, 1.0, 0.7, 0.4], dtype=xp.float64))

    @make_xp_test_case(windows.lanczos)
    def test_lanczos(self, xp):
        xp_assert_close(get_window('lanczos', 6, xp=xp),
                        xp.asarray([0., 0.413496672, 0.826993343, 1., 0.826993343,
                                    0.413496672], dtype=xp.float64), atol=1e-9)
        xp_assert_close(get_window('lanczos', 6, fftbins=False, xp=xp),
                        xp.asarray([0., 0.504551152, 0.935489284, 0.935489284,
                                    0.504551152, 0.], dtype=xp.float64), atol=1e-9)
        xp_assert_close(get_window('lanczos', 6, xp=xp),
                        get_window('sinc', 6, xp=xp))

    def test_xp_default(self, xp):
        # no explicit xp= argument, default to numpy
        win = get_window('lanczos', 6)
        assert isinstance(win, np.ndarray)

        win = get_window('lanczos', 6, xp=xp)
        if not is_numpy(xp):
            assert not isinstance(win, np.ndarray)


@make_xp_test_case(windows.general_gaussian)
class TestGeneralGaussian:

    @pytest.mark.parametrize('M', (6, 7))
    @pytest.mark.parametrize('p', (1, 1.5))
    @pytest.mark.parametrize('sig', (1., 2.))
    @pytest.mark.parametrize('sym', (True, False))
    def test_basic(self, M, p, sig, sym, xp):
        """Verify basic parametrizations. """
        if sym:
            t = xp.arange(M, dtype=xp.float64) - (M-1) / 2
        else:
            t = xp.arange(M+1, dtype=xp.float64) - M / 2
            t = t[:-1]
        x_ref = xp.exp(-0.5 * abs(t / sig) ** (2*p))
        x = windows.general_gaussian(M, p, sig, sym, xp=xp)
        xp_assert_close(x, x_ref, atol=1e-12)

    def test_default_func_parameter(self, xp):
        """Verify the correctness of the default value of function parameter `sym`. """
        x = windows.general_gaussian(6, 1, 1, xp=xp)
        x_ref = windows.general_gaussian(6, 1, 1, sym=True, xp=xp)
        xp_assert_equal(x, x_ref)

    def test_p1_equals_gaussian(self,xp):
        """Testing that `general_gaussian` with p = 1 is equivalent to the normal
        `gaussian` window. """
        w1 = windows.general_gaussian(7, p=1.0, sig=2.0, xp=xp)
        w2 = windows.gaussian(7, std=2.0, xp=xp)
        xp_assert_close(w1, w2)

    def test_peak_value(self, xp):
        """ Testing that if M is odd, the peak is at 1. """
        w = windows.general_gaussian(7, p=3, sig=2.0, xp=xp)
        assert float(xp.max(w)) == 1.0

        # Testing that if M is even, the peak is less than 1.0
        w = windows.general_gaussian(6, p=1.5, sig=2.0, xp=xp)
        assert float(xp.max(w)) < 1.0

    def test_len_edge_cases(self, xp):
        """Test the length edge cases are handled correctly. """
        # length = 0 should return an empty array:
        assert windows.general_gaussian(0, 1, 1, xp=xp).shape[0] == 0

        # length = 1 should return an array of length 1 containing 1.0:
        xp_assert_close(windows.general_gaussian(1, 1, 1, xp=xp),
                        xp.asarray([1.0], dtype=xp.float64))

        with pytest.raises(ValueError):  # Negative values should raise an error:
            windows.general_gaussian(-1, 1, 1, xp=xp)


@make_xp_test_case(windows.cosine)
class TestCosine:
    @pytest.mark.parametrize('M', (3, 4))
    @pytest.mark.parametrize('sym', (True, False))
    def test_basic(self, M, sym, xp):
        """Verify basic parametrizations. """
        if sym:
            t = (xp.arange(M, dtype=xp.float64) + 0.5) / M
        else:
            t = (xp.arange(M+1, dtype=xp.float64) + 0.5) / (M+1)
            t = t[:-1]
        x_ref = xp.sin(xp.pi * t)
        x = windows.cosine(M, sym=sym, xp=xp)
        xp_assert_close(x, x_ref, atol=1e-12)

    def test_default_func_parameter(self, xp):
        """Verify the correctness of the default value of function parameter `sym`. """
        x = windows.cosine(6, xp=xp)
        x_ref = windows.cosine(6, sym=True, xp=xp)
        xp_assert_equal(x, x_ref)

    def test_len_edge_cases(self, xp):
        """Testing that the length edge cases are handled correctly."""
        # length = 0 should return an empty array:
        assert windows.cosine(0, xp=xp).shape[0] == 0

        # length = 1 should return an array of length 1 containing 1.0:
        xp_assert_close(windows.cosine(1, xp=xp),
                        xp.asarray([1.0], dtype=xp.float64))

        with pytest.raises(ValueError):
            windows.cosine(-1, xp=xp) # negative values should return an error


@skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/2620")
@pytest.mark.parametrize(
    "window,window_name,params",
    [
        make_xp_pytest_param(getattr(windows, window_name), window_name, params)
        for window_name, params in window_funcs
    ]
)
def test_windowfunc_basics(window, window_name, params, xp):
    window = getattr(windows, window_name)
    if is_jax(xp) and window_name in ['taylor', 'chebwin']:
        pytest.skip(reason=f'{window_name = }: item assignment')
    if window_name in ['dpss']:
        if is_cupy(xp):
            pytest.skip(reason='dpss window is not implemented for cupy')
        if is_torch(xp) and SCIPY_DEVICE != 'cpu':
            pytest.skip(reason='needs eight_tridiagonal which is CPU only')

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", "This window is not suitable", UserWarning)
        # Check symmetry for odd and even lengths
        w1 = window(8, *params, sym=True, xp=xp)
        w2 = window(7, *params, sym=False, xp=xp)
        xp_assert_close(w1[:-1], w2)

        w1 = window(9, *params, sym=True, xp=xp)
        w2 = window(8, *params, sym=False, xp=xp)
        xp_assert_close(w1[:-1], w2)

        # Check that functions run and output lengths are correct
        assert window(6, *params, sym=True, xp=xp).shape[0] == 6
        assert window(6, *params, sym=False, xp=xp).shape[0] == 6
        assert window(7, *params, sym=True, xp=xp).shape[0] == 7
        assert window(7, *params, sym=False, xp=xp).shape[0] == 7

        # Check invalid lengths
        assert_raises(ValueError, window, 5.5, *params, xp=xp)
        assert_raises(ValueError, window, -7, *params, xp=xp)

        # Check degenerate cases
        xp_assert_equal(window(0, *params, sym=True, xp=xp),
                        xp.asarray([], dtype=xp.float64))
        xp_assert_equal(window(0, *params, sym=False, xp=xp),
                        xp.asarray([], dtype=xp.float64))
        xp_assert_equal(window(1, *params, sym=True, xp=xp),
                        xp.asarray([1.], dtype=xp.float64))
        xp_assert_equal(window(1, *params, sym=False, xp=xp),
                        xp.asarray([1.], dtype=xp.float64))

        # Check dtype
        assert window(0, *params, sym=True, xp=xp).dtype == xp.float64
        assert window(0, *params, sym=False, xp=xp).dtype == xp.float64
        assert window(1, *params, sym=True, xp=xp).dtype == xp.float64
        assert window(1, *params, sym=False, xp=xp).dtype == xp.float64
        assert window(6, *params, sym=True, xp=xp).dtype == xp.float64
        assert window(6, *params, sym=False, xp=xp).dtype == xp.float64

        # Check normalization
        assert xp.all(window(10, *params, sym=True, xp=xp) < 1.01)
        assert xp.all(window(10, *params, sym=False, xp=xp) < 1.01)
        assert xp.all(window(9, *params, sym=True, xp=xp) < 1.01)
        assert xp.all(window(9, *params, sym=False, xp=xp) < 1.01)

        # Check that DFT-even spectrum is purely real for odd and even
        res = fft(window(10, *params, sym=False, xp=xp))
        res = xp.imag(res)
        xp_assert_close(res, xp.zeros_like(res), atol=1e-14)

        res = fft(window(11, *params, sym=False, xp=xp))
        res = xp.imag(res)
        xp_assert_close(res, xp.zeros_like(res), atol=1e-14)


@make_xp_test_case(get_window)
def test_needs_params(xp):
    for winstr in ['kaiser', 'ksr', 'kaiser_bessel_derived', 'kbd',
                   'gaussian', 'gauss', 'gss',
                   'general gaussian', 'general_gaussian',
                   'general gauss', 'general_gauss', 'ggs',
                   'dss', 'dpss', 'general cosine', 'general_cosine',
                   'chebwin', 'cheb', 'general hamming', 'general_hamming',
                   ]:
        assert_raises(ValueError, get_window, winstr, 7, xp=xp)


_winstr = ['barthann',
           'bartlett',
           'blackman',
           'blackmanharris',
           'bohman',
           'boxcar',
           'cosine',
           'flattop',
           'hamming',
           'nuttall',
           'parzen',
           'taylor',
           'exponential',
           'poisson',
           'tukey',
           'tuk',
           'triangle',
           'lanczos',
           'sinc',
]


@pytest.mark.parametrize(
    'window,winstr',
    [
        make_xp_pytest_param(_WIN_FUNCS[winstr][0], winstr)
        for winstr in _winstr
    ]
)
@make_xp_test_case(get_window)
def test_not_needs_params(xp, window, winstr):
    if is_jax(xp) and winstr in ['taylor']:
        pytest.skip(reason=f'{winstr}: item assignment')
    win = get_window(winstr, 7, xp=xp)
    assert win.shape[0] == 7


@make_xp_test_case(windows.lanczos)
def test_symmetric(xp):

    for win in [windows.lanczos]:
        # Even sampling points
        w = win(4096, xp=xp)
        flip = array_namespace(w).flip
        error = xp.max(xp.abs(w - flip(w)))
        xp_assert_equal(error, xp.asarray(0.0), check_dtype=False, check_0d=False)

        # Odd sampling points
        w = win(4097, xp=xp)
        error = xp.max(xp.abs(w - flip(w)))
        xp_assert_equal(error, xp.asarray(0.0), check_dtype=False, check_0d=False)



from enum import Enum
from types import ModuleType
from typing import (
    Final,
    Literal as L,
    NotRequired,
    TypedDict,
    overload,
    type_check_only,
)

_CompilerConfigDictValue = TypedDict(
    "_CompilerConfigDictValue",
    {
        "name": str,
        "linker": str,
        "version": str,
        "commands": str,
        "args": str,
        "linker args": str,
    },
)
_CompilerConfigDict = TypedDict(
    "_CompilerConfigDict",
    {
        "c": _CompilerConfigDictValue,
        "cython": _CompilerConfigDictValue,
        "c++": _CompilerConfigDictValue,
    },
)
_MachineInformationDict = TypedDict(
    "_MachineInformationDict",
    {
        "host": _MachineInformationDictValue,
        "build": _MachineInformationDictValue,
        "cross-compiled": NotRequired[L[True]],
    },
)

@type_check_only
class _MachineInformationDictValue(TypedDict):
    cpu: str
    family: str
    endian: L["little", "big"]
    system: str

_BuildDependenciesDictValue = TypedDict(
    "_BuildDependenciesDictValue",
    {
        "name": str,
        "found": NotRequired[L[True]],
        "version": str,
        "include directory": str,
        "lib directory": str,
        "openblas configuration": str,
        "pc file directory": str,
    },
)

class _BuildDependenciesDict(TypedDict):
    blas: _BuildDependenciesDictValue
    lapack: _BuildDependenciesDictValue

class _PythonInformationDict(TypedDict):
    path: str
    version: str

_SIMDExtensionsDict = TypedDict(
    "_SIMDExtensionsDict",
    {
        "baseline": list[str],
        "found": list[str],
        "not found": list[str],
    },
)

_ConfigDict = TypedDict(
    "_ConfigDict",
    {
        "Compilers": _CompilerConfigDict,
        "Machine Information": _MachineInformationDict,
        "Build Dependencies": _BuildDependenciesDict,
        "Python Information": _PythonInformationDict,
        "SIMD Extensions": _SIMDExtensionsDict,
    },
)

###

__all__ = ["show_config"]

CONFIG: Final[_ConfigDict] = ...

class DisplayModes(Enum):
    stdout = "stdout"
    dicts = "dicts"

def _check_pyyaml() -> ModuleType: ...

@overload
def show(mode: L["stdout"] = "stdout") -> None: ...
@overload
def show(mode: L["dicts"]) -> _ConfigDict: ...

@overload
def show_config(mode: L["stdout"] = "stdout") -> None: ...
@overload
def show_config(mode: L["dicts"]) -> _ConfigDict: ...



from typing import Any

import numpy as np

AR_U: np.char.chararray[tuple[Any, ...], np.dtype[np.str_]]  # type: ignore[deprecated]
AR_S: np.char.chararray[tuple[Any, ...], np.dtype[np.bytes_]]  # type: ignore[deprecated]

AR_S.encode()  # type: ignore[misc]
AR_U.decode()  # type: ignore[misc]

AR_U.join(b"_")  # type: ignore[arg-type]
AR_S.join("_")  # type: ignore[arg-type]

AR_U.ljust(5, fillchar=b"a")  # type: ignore[arg-type]
AR_U.rjust(5, fillchar=b"a")  # type: ignore[arg-type]

AR_U.lstrip(chars=b"a")  # type: ignore[arg-type]
AR_S.lstrip(chars="a")  # type: ignore[arg-type]
AR_U.strip(chars=b"a")  # type: ignore[arg-type]
AR_S.strip(chars="a")  # type: ignore[arg-type]
AR_U.rstrip(chars=b"a")  # type: ignore[arg-type]
AR_S.rstrip(chars="a")  # type: ignore[arg-type]

AR_U.partition(b"a")  # type: ignore[arg-type]
AR_S.partition("a")  # type: ignore[arg-type]
AR_U.rpartition(b"a")  # type: ignore[arg-type]
AR_S.rpartition("a")  # type: ignore[arg-type]

AR_U.replace(b"_", b"-")  # type: ignore[arg-type]
AR_S.replace("_", "-")  # type: ignore[arg-type]

AR_U.split(b"_")  # type: ignore[arg-type]
AR_S.split("_")  # type: ignore[arg-type]
AR_S.split(1)  # type: ignore[arg-type]
AR_U.rsplit(b"_")  # type: ignore[arg-type]
AR_S.rsplit("_")  # type: ignore[arg-type]

AR_U.count(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.count("a", end=9)  # type: ignore[arg-type]

AR_U.endswith(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.endswith("a", end=9)  # type: ignore[arg-type]
AR_U.startswith(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.startswith("a", end=9)  # type: ignore[arg-type]

AR_U.find(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.find("a", end=9)  # type: ignore[arg-type]
AR_U.rfind(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.rfind("a", end=9)  # type: ignore[arg-type]

AR_U.index(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.index("a", end=9)  # type: ignore[arg-type]
AR_U.rindex(b"a", start=[1, 2, 3])  # type: ignore[arg-type]
AR_S.rindex("a", end=9)  # type: ignore[arg-type]

AR_U == AR_S  # type: ignore[operator]
AR_U != AR_S  # type: ignore[operator]
AR_U >= AR_S  # type: ignore[operator]
AR_U <= AR_S  # type: ignore[operator]
AR_U > AR_S  # type: ignore[operator]
AR_U < AR_S  # type: ignore[operator]



'''
This is a module that preprocesses the inputs.

Translated from Zaikun Zhang's modern-Fortran reference implementation in PRIMA.

Dedicated to late Professor M. J. D. Powell FRS (1936--2015).

Python translation by Nickolai Belakovski.
'''

from .consts import DEBUGGING, EPS, IPRINT_DEFAULT, FTARGET_DEFAULT, \
    MIN_MAXFILT, MAXFILT_DEFAULT, MAXHISTMEM, ETA1_DEFAULT, ETA2_DEFAULT, \
    GAMMA1_DEFAULT, GAMMA2_DEFAULT, RHOBEG_DEFAULT, RHOEND_DEFAULT, \
    CTOL_DEFAULT, CWEIGHT_DEFAULT
from .present import present
from warnings import warn
import numpy as np


def preproc(solver, num_vars, iprint, maxfun, maxhist, ftarget, rhobeg, rhoend,
            num_constraints=None, npt=None, maxfilt=None, ctol=None, cweight=None,
            eta1=None, eta2=None, gamma1=None, gamma2=None, is_constrained=None, has_rhobeg=None,
            honour_x0=None, xl=None, xu=None, x0=None):
    '''
    This subroutine preprocesses the inputs. It does nothing to the inputs that are valid.
    '''
    # Preconditions
    if DEBUGGING:
        assert num_vars >= 1
        if present(num_constraints):
            assert num_constraints >= 0
            assert num_constraints == 0 or solver.lower() == 'cobyla'
        if solver.lower() == 'cobyla' and present(num_constraints) and present(is_constrained):
            assert num_constraints == 0 or is_constrained
        if solver.lower() == 'bobyqa':
            assert present(xl) and present(xu)
            assert all(xu - xl >= 2 * EPS)
        if present(honour_x0):
            assert solver.lower() == 'bobyqa' and present(has_rhobeg) and present(xl) and present(xu) and present(x0)

    #====================#
    # Calculation starts #
    #====================#

    # Read num_constraints, if necessary
    num_constraints = num_constraints if (present(num_constraints) and solver.lower() == 'cobyla') else 0

    # Decide whether the problem is truly constrained
    is_constrained = is_constrained if (present(is_constrained)) else num_constraints > 0

    # Validate IPRINT
    if np.abs(iprint) > 3:
        iprint = IPRINT_DEFAULT
        warn(f'{solver}: Invalid IPRINT; it should be 0, 1, -1, 2, -2, 3, or -3; it is set to {iprint}')

    # Validate MAXFUN
    if solver.lower() == 'uobyqa':
        min_maxfun = (num_vars + 1) * (num_vars + 2) / 2 + 1
        min_maxfun_str = '(N+1)(N+2)/2 + 1'
    elif solver.lower() == 'cobyla':
        min_maxfun = num_vars + 2
        min_maxfun_str = 'num_vars + 2'
    else:  # CASE ('NEWUOA', 'BOBYQA', 'LINCOA')
        min_maxfun = num_vars + 3
        min_maxfun_str = 'num_vars + 3'
    if maxfun < min_maxfun:
        maxfun = min_maxfun
        warn(f'{solver}: Invalid MAXFUN; it should be at least {min_maxfun_str}; it is set to {maxfun}')

    # Validate MAXHIST
    if maxhist < 0:
        maxhist = maxfun
        warn(f'{solver}: Invalid MAXHIST; it should be a nonnegative integer; it is set to {maxhist}')
    maxhist = min(maxhist, maxfun)  # MAXHIST > MAXFUN is never needed.

    # Validate FTARGET
    if np.isnan(ftarget):
        ftarget = FTARGET_DEFAULT
        warn(f'{solver}: Invalid FTARGET; it should be a real number; it is set to {ftarget}')

    # Validate NPT
    if (solver.lower() == 'newuoa' or solver.lower() == 'bobyqa' or solver.lower() == 'lincoa') and present(npt):
        if (npt < num_vars + 2 or npt > min(maxfun - 1, ((num_vars + 2) * (num_vars + 1)) / 2)):
            npt = int(min(maxfun - 1, 2 * num_vars + 1))
            warn(f'{solver}: Invalid NPT; it should be an integer in the interval [N+2, (N+1)(N+2)/2] and less than MAXFUN; it is set to {npt}')

    # Validate MAXFILT
    if present(maxfilt) and (solver.lower() == 'lincoa' or solver.lower() == 'cobyla'):
        maxfilt_in = maxfilt
        if maxfilt < 1:
            maxfilt = MAXFILT_DEFAULT  # The inputted MAXFILT is obviously wrong.
        else:
            maxfilt = max(MIN_MAXFILT, maxfilt)  # The inputted MAXFILT is too small.
        # Further revise MAXFILT according to MAXHISTMEM.
        if solver.lower() == 'lincoa':
            unit_memo = (num_vars + 2) * np.dtype(float).itemsize
        elif solver.lower() == 'cobyla':
            unit_memo = (num_constraints + num_vars + 2) * np.dtype(float).itemsize
        else:
            unit_memo = 1
        # We cannot simply set MAXFILT = MIN(MAXFILT, MAXHISTMEM/...), as they may not have
        # the same kind, and compilers may complain. We may convert them, but overflow may occur.
        if maxfilt > MAXHISTMEM / unit_memo:
            maxfilt = int(MAXHISTMEM / unit_memo)  # Integer division.
        maxfilt = min(maxfun, max(MIN_MAXFILT, maxfilt))
        if is_constrained:
            if maxfilt_in < 1:
                warn(f'{solver}: Invalid MAXFILT; it should be a positive integer; it is set to  {maxfilt}')
            elif maxfilt_in < min(maxfun, MIN_MAXFILT):
                warn(f'{solver}: MAXFILT is too small; it is set to {maxfilt}')
            elif maxfilt < min(maxfilt_in, maxfun):
                warn(f'{solver}: MAXFILT is set to {maxfilt} due to memory limit')

    # Validate ETA1 and ETA2
    eta1_local = eta1 if present(eta1) else ETA1_DEFAULT
    eta2_local = eta2 if present(eta2) else ETA2_DEFAULT

    # When the difference between ETA1 and ETA2 is tiny, we force them to equal.
    # See the explanation around RHOBEG and RHOEND for the reason.
    if present(eta1) and present(eta2):
        if np.abs(eta1 - eta2) < 1.0E2 * EPS * max(np.abs(eta1), 1):
            eta2 = eta1

    if present(eta1):
        if np.isnan(eta1):
            # In this case, we take the value hard coded in Powell's original code
            # without any warning. It is useful when interfacing with MATLAB/Python.
            eta1 = ETA1_DEFAULT
        elif eta1 < 0 or eta1 >= 1:
            # Take ETA2 into account if it has a valid value.
            if present(eta2) and eta2_local > 0 and eta2_local <= 1:
                eta1 = max(EPS, eta2 / 7.0)
            else:
                eta1 = ETA1_DEFAULT
            warn(f'{solver}: Invalid ETA1; it should be in the interval [0, 1) and not more than ETA2; it is set to {eta1}')

    if present(eta2):
        if np.isnan(eta2):
            # In this case, we take the value hard coded in Powell's original code
            # without any warning. It is useful when interfacing with MATLAB/Python.
            eta2 = ETA2_DEFAULT
        elif present(eta1) and (eta2 < eta1_local or eta2 > 1):
            eta2 = (eta1 + 2) / 3.0
            warn(f'{solver}: Invalid ETA2; it should be in the interval [0, 1) and not less than ETA1; it is set to {eta2}')

    # Validate GAMMA1 and GAMMA2
    if present(gamma1):
        if np.isnan(gamma1):
            # In this case, we take the value hard coded in Powell's original code
            # without any warning. It is useful when interfacing with MATLAB/Python.
            gamma1 = GAMMA1_DEFAULT
        elif gamma1 <= 0 or gamma1 >= 1:
            gamma1 = GAMMA1_DEFAULT
            warn(f'{solver}: Invalid GAMMA1; it should in the interval (0, 1); it is set to {gamma1}')

    if present(gamma2):
        if np.isnan(gamma2):
            # In this case, we take the value hard coded in Powell's original code
            # without any warning. It is useful when interfacing with MATLAB/Python.
            gamma2 = GAMMA2_DEFAULT
        elif gamma2 < 1 or np.isinf(gamma2):
            gamma2 = GAMMA2_DEFAULT
            warn(f'{solver}: Invalid GAMMA2; it should be a real number not less than 1; it is set to {gamma2}')

    # Validate RHOBEG and RHOEND

    if np.abs(rhobeg - rhoend) < 1.0e2 * EPS * np.maximum(np.abs(rhobeg), 1):
        # When the data is passed from the interfaces (e.g., MEX) to the Fortran code, RHOBEG, and RHOEND
        # may change a bit. It was observed in a MATLAB test that MEX passed 1 to Fortran as
        # 0.99999999999999978. Therefore, if we set RHOEND = RHOBEG in the interfaces, then it may happen
        # that RHOEND > RHOBEG, which is considered as an invalid input. To avoid this situation, we
        # force RHOBEG and RHOEND to equal when the difference is tiny.
        rhoend = rhobeg

    # Revise the default values for RHOBEG/RHOEND according to the solver.
    if solver.lower() == 'bobyqa':
        rhobeg_default = np.maximum(EPS, np.min(RHOBEG_DEFAULT, np.min(xu - xl) / 4.0))
        rhoend_default = np.maximum(EPS, np.min(0.1 * rhobeg_default, RHOEND_DEFAULT))
    else:
        rhobeg_default = RHOBEG_DEFAULT
        rhoend_default = RHOEND_DEFAULT

    if solver.lower() == 'bobyqa':
        # Do NOT merge the IF below into the ELIF above! Otherwise, XU and XL may be accessed even if
        # the solver is not BOBYQA, because the logical evaluation is not short-circuit.
        if rhobeg > np.min(xu - xl) / 2:
            # Do NOT make this revision if RHOBEG not positive or not finite, because otherwise RHOBEG
            # will get a huge value when XU or XL contains huge values that indicate unbounded variables.
            rhobeg = np.min(xu - xl) / 4.0  # Here, we do not take RHOBEG_DEFAULT.
            warn(f'{solver}: Invalid RHOBEG; {solver} requires 0 < RHOBEG <= np.min(XU-XL)/2; it is set to np.min(XU-XL)/4')
    if rhobeg <= 0 or np.isnan(rhobeg) or np.isinf(rhobeg):
        # Take RHOEND into account if it has a valid value. We do not do this if the solver is BOBYQA,
        # which requires that RHOBEG <= (XU-XL)/2.
        if np.isfinite(rhoend) and rhoend > 0 and solver.lower() != 'bobyqa':
            rhobeg = max(10 * rhoend, rhobeg_default)
        else:
            rhobeg = rhobeg_default
        warn(f'{solver}: Invalid RHOBEG; it should be a positive number; it is set to {rhobeg}')

    if rhoend <= 0 or rhobeg < rhoend or np.isnan(rhoend) or np.isinf(rhoend):
        rhoend = max(EPS, min(0.1 * rhobeg, rhoend_default))
        warn(f'{solver}: Invalid RHOEND; it should be a positive number and RHOEND <= RHOBEG; it is set to {rhoend}')

    # For BOBYQA, revise X0 or RHOBEG so that the distance between X0 and the inactive bounds is at
    # least RHOBEG. If HONOUR_X0 == TRUE, revise RHOBEG if needed; otherwise, revise HONOUR_X0 if needed.
    if present(honour_x0):
        if honour_x0:
            rhobeg_old = rhobeg;
            lbx = np.isfinite(xl) & (x0 - xl <= EPS * np.maximum(1, np.abs(xl))) # X0 essentially equals XL
            ubx = np.isfinite(xu) & (x0 - xu >= -EPS * np.maximum(1, np.abs(xu))) # X0 essentially equals XU
            x0[lbx] = xl[lbx]
            x0[ubx] = xu[ubx]
            rhobeg = max(EPS, np.min([rhobeg, x0[~lbx] - xl[~lbx], xu[~ubx] - x0[~ubx]]))
            if rhobeg_old - rhobeg > EPS * max(1, rhobeg_old):
                rhoend = max(EPS, min(0.1 * rhobeg, rhoend)) # We do not revise RHOEND unless RHOBEG is truly revised.
                if has_rhobeg:
                    warn(f'{solver}: RHOBEG is revised to {rhobeg} and RHOEND to at most 0.1*RHOBEG so that the distance between X0 and the inactive bounds is at least RHOBEG')
            else:
                rhoend = np.minimum(rhoend, rhobeg)  # This may update RHOEND slightly.
        else:
            x0_old = x0  # Recorded to see whether X0 is really revised.
            # N.B.: The following revision is valid only if XL <= X0 <= XU and RHOBEG <= MINVAL(XU-XL)/2,
            # which should hold at this point due to the revision of RHOBEG and moderation of X0.
            # The cases below are mutually exclusive in precise arithmetic as MINVAL(XU-XL) >= 2*RHOBEG.
            lbx = x0 <= xl + 0.5 * rhobeg
            lbx_plus = (x0 > xl + 0.5 * rhobeg) & (x0 < xl + rhobeg)
            ubx = x0 >= xu - 0.5 * rhobeg
            ubx_minus = (x0 < xu - 0.5 * rhobeg) & (x0 > xu - rhobeg)
            x0[lbx] = xl[lbx]
            x0[lbx_plus] = xl[lbx_plus] + rhobeg
            x0[ubx] = xu[ubx]
            x0[ubx_minus] = xu[ubx_minus] - rhobeg

            if (any(np.abs(x0_old - x0) > 0)):
                warn(f'{solver}: X0 is revised so that the distance between X0 and the inactive bounds is at least RHOBEG set HONOUR_X0 to .TRUE. if you prefer to keep X0 unchanged')

    # Validate CTOL (it can be 0)
    if (present(ctol)):
        if (np.isnan(ctol) or ctol < 0):
            ctol = CTOL_DEFAULT
            if (is_constrained):
                warn(f'{solver}: Invalid CTOL; it should be a nonnegative number; it is set to {ctol}')

    # Validate CWEIGHT (it can be +Inf)
    if (present(cweight)):
        if (np.isnan(cweight) or cweight < 0):
            cweight = CWEIGHT_DEFAULT
            if (is_constrained):
                warn(f'{solver}: Invalid CWEIGHT; it should be a nonnegative number; it is set to {cweight}')

    #====================#
    #  Calculation ends  #
    #====================#

    # Postconditions
    if DEBUGGING:
        assert abs(iprint) <= 3
        assert maxhist >= 0 and maxhist <= maxfun
        if present(npt):
            assert maxfun >= npt + 1
            assert npt >= 3
        if present(maxfilt):
            assert maxfilt >= np.minimum(MIN_MAXFILT, maxfun) and maxfilt <= maxfun
        if present(eta1) and present(eta2):
            assert eta1 >= 0 and eta1 <= eta2 and eta2 < 1
        if present(gamma1) and present(gamma2):
            assert gamma1 > 0 and gamma1 < 1 and gamma2 > 1
        assert rhobeg >= rhoend and rhoend > 0
        if solver.lower() == 'bobyqa':
            assert all(rhobeg <= (xu - xl) / 2)
            assert all(np.isfinite(x0))
            assert all(x0 >= xl and (x0 <= xl or x0 >= xl + rhobeg))
            assert all(x0 <= xu and (x0 >= xu or x0 <= xu - rhobeg))
        if present(ctol):
            assert ctol >= 0

    return iprint, maxfun, maxhist, ftarget, rhobeg, rhoend, npt, maxfilt, ctol, cweight, eta1, eta2, gamma1, gamma2, x0



from collections.abc import Container, Iterable
from typing import Any, Literal as L, Protocol, overload, type_check_only
from typing_extensions import deprecated

import numpy as np
from numpy._typing import (
    ArrayLike,
    NDArray,
    _16Bit,
    _32Bit,
    _64Bit,
    _AnyShape,
    _ArrayLike,
    _NestedSequence,
    _ScalarLike_co,
    _Shape,
    _SupportsArray,
)

__all__ = [
    "common_type",
    "imag",
    "iscomplex",
    "iscomplexobj",
    "isreal",
    "isrealobj",
    "mintypecode",
    "nan_to_num",
    "real",
    "real_if_close",
    "typename",
]

type _FloatMax32 = np.float32 | np.float16
type _ComplexMax128 = np.complex128 | np.complex64
type _RealMax64 = np.float64 | np.float32 | np.float16 | np.integer
type _Real = np.floating | np.integer
type _ToReal = _Real | np.bool
type _InexactMax32 = np.inexact[_32Bit] | np.float16
type _NumberMax64 = np.number[_64Bit] | np.number[_32Bit] | np.number[_16Bit] | np.integer

@type_check_only
class _HasReal[T](Protocol):
    @property
    def real(self, /) -> T: ...

@type_check_only
class _HasImag[T](Protocol):
    @property
    def imag(self, /) -> T: ...

@type_check_only
class _HasDType[ScalarT: np.generic](Protocol):
    @property
    def dtype(self, /) -> np.dtype[ScalarT]: ...

###

def mintypecode(typechars: Iterable[str | ArrayLike], typeset: str | Container[str] = "GDFgdf", default: str = "d") -> str: ...

#
@overload
def real[T](val: _HasReal[T]) -> T: ...
@overload
def real[RealT: _ToReal](val: _ArrayLike[RealT]) -> NDArray[RealT]: ...
@overload
def real(val: ArrayLike) -> NDArray[Any]: ...

#
@overload
def imag[T](val: _HasImag[T]) -> T: ...
@overload
def imag[RealT: _ToReal](val: _ArrayLike[RealT]) -> NDArray[RealT]: ...
@overload
def imag(val: ArrayLike) -> NDArray[Any]: ...

#
@overload
def iscomplex(x: _ScalarLike_co) -> np.bool: ...
@overload
def iscomplex(x: NDArray[Any] | _NestedSequence[ArrayLike]) -> NDArray[np.bool]: ...
@overload
def iscomplex(x: ArrayLike) -> np.bool | NDArray[np.bool]: ...

#
@overload
def isreal(x: _ScalarLike_co) -> np.bool: ...
@overload
def isreal(x: NDArray[Any] | _NestedSequence[ArrayLike]) -> NDArray[np.bool]: ...
@overload
def isreal(x: ArrayLike) -> np.bool | NDArray[np.bool]: ...

#
def iscomplexobj(x: _HasDType[Any] | ArrayLike) -> bool: ...
def isrealobj(x: _HasDType[Any] | ArrayLike) -> bool: ...

#
@overload  # np.generic | np.ndarray  (`ndarray` subclasses pass through)
def nan_to_num[ScalarOrArrayT: np.generic | np.ndarray](
    x: ScalarOrArrayT,
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> ScalarOrArrayT: ...
@overload  # >0-d <known dtype>
def nan_to_num[ScalarT: np.generic](
    x: _NestedSequence[_ArrayLike[ScalarT]],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?-d <known dtype>
def nan_to_num[DTypeT: np.dtype](
    x: _SupportsArray[DTypeT],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.ndarray[_AnyShape, DTypeT] | Any: ...
@overload  # 0-d ~bool
def nan_to_num(
    x: bool,
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.bool: ...
@overload  # 0-d +int
def nan_to_num(
    x: int,
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.int_ | Any: ...
@overload  # 0-d +float
def nan_to_num(
    x: float,
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.float64 | Any: ...
@overload  # 0-d +complex
def nan_to_num(
    x: complex,
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.complex128 | Any: ...
@overload  # >0-d ~bool
def nan_to_num(
    x: _NestedSequence[bool],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> NDArray[np.bool]: ...
@overload  # >0-d ~int
def nan_to_num(
    x: _NestedSequence[list[int]] | list[int],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> NDArray[np.int_]: ...
@overload  # >0-d ~float
def nan_to_num(
    x: _NestedSequence[list[float]] | list[float],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> NDArray[np.float64]: ...
@overload  # >0-d ~complex
def nan_to_num(
    x: _NestedSequence[list[complex]] | list[complex],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> NDArray[np.complex128]: ...
@overload  # >0-d <unknown dtype>
def nan_to_num(
    x: _NestedSequence[ArrayLike],
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.ndarray: ...
@overload  # ?-d <unknown dtype>
def nan_to_num(
    x: ArrayLike,
    copy: bool = True,
    nan: float = 0.0,
    posinf: float | None = None,
    neginf: float | None = None,
) -> np.ndarray | Any: ...

#
@overload
def real_if_close[ShapeT: _Shape, DTypeT: np.dtype[_ToReal]](
    a: np.ndarray[ShapeT, DTypeT],
    tol: float = 100,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload
def real_if_close[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex64]],
    tol: float = 100,
) -> np.ndarray[ShapeT, np.dtype[np.float32 | np.complex64]]: ...
@overload
def real_if_close[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex128]],
    tol: float = 100,
) -> np.ndarray[ShapeT, np.dtype[np.float64 | np.complex128]]: ...
@overload
def real_if_close[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.clongdouble]],
    tol: float = 100,
) -> np.ndarray[ShapeT, np.dtype[np.longdouble | np.clongdouble]]: ...
@overload
def real_if_close[RealT: _ToReal](a: _ArrayLike[RealT], tol: float = 100) -> NDArray[RealT]: ...
@overload
def real_if_close(a: _ArrayLike[np.complex64], tol: float = 100) -> NDArray[np.float32 | np.complex64]: ...
@overload
def real_if_close(a: _ArrayLike[np.complex128], tol: float = 100) -> NDArray[np.float64 | np.complex128]: ...
@overload
def real_if_close(a: _ArrayLike[np.clongdouble], tol: float = 100) -> NDArray[np.longdouble | np.clongdouble]: ...
@overload
def real_if_close(a: ArrayLike, tol: float = 100) -> NDArray[Any]: ...

#
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["S1"]) -> L["character"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["?"]) -> L["bool"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["b"]) -> L["signed char"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["B"]) -> L["unsigned char"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["h"]) -> L["short"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["H"]) -> L["unsigned short"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["i"]) -> L["integer"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["I"]) -> L["unsigned integer"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["l"]) -> L["long integer"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["L"]) -> L["unsigned long integer"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["q"]) -> L["long long integer"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["Q"]) -> L["unsigned long long integer"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["f"]) -> L["single precision"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["d"]) -> L["double precision"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["g"]) -> L["long precision"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["F"]) -> L["complex single precision"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["D"]) -> L["complex double precision"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["G"]) -> L["complex long double precision"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["S"]) -> L["string"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["U"]) -> L["unicode"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["V"]) -> L["void"]: ...
@overload
@deprecated("numpy.typename is deprecated. Use numpy.dtype.name instead.")
def typename(char: L["O"]) -> L["object"]: ...

# NOTE: The [overload-overlap] mypy errors are false positives
@overload
def common_type() -> type[np.float16]: ...
@overload
def common_type(a0: _HasDType[np.float16], /, *ai: _HasDType[np.float16]) -> type[np.float16]: ...
@overload
def common_type(a0: _HasDType[np.float32], /, *ai: _HasDType[_FloatMax32]) -> type[np.float32]: ...
@overload
def common_type(
    a0: _HasDType[np.float64 | np.integer],
    /,
    *ai: _HasDType[_RealMax64],
) -> type[np.float64]: ...
@overload
def common_type(
    a0: _HasDType[np.longdouble],
    /,
    *ai: _HasDType[_Real],
) -> type[np.longdouble]: ...
@overload
def common_type(
    a0: _HasDType[np.complex64],
    /,
    *ai: _HasDType[_InexactMax32],
) -> type[np.complex64]: ...
@overload
def common_type(
    a0: _HasDType[np.complex128],
    /,
    *ai: _HasDType[_NumberMax64],
) -> type[np.complex128]: ...
@overload
def common_type(
    a0: _HasDType[np.clongdouble],
    /,
    *ai: _HasDType[np.number],
) -> type[np.clongdouble]: ...
@overload
def common_type(
    a0: _HasDType[_FloatMax32],
    array1: _HasDType[np.float32],
    /,
    *ai: _HasDType[_FloatMax32],
) -> type[np.float32]: ...
@overload
def common_type(
    a0: _HasDType[_RealMax64],
    array1: _HasDType[np.float64 | np.integer],
    /,
    *ai: _HasDType[_RealMax64],
) -> type[np.float64]: ...
@overload
def common_type(
    a0: _HasDType[_Real],
    array1: _HasDType[np.longdouble],
    /,
    *ai: _HasDType[_Real],
) -> type[np.longdouble]: ...
@overload
def common_type(
    a0: _HasDType[_InexactMax32],
    array1: _HasDType[np.complex64],
    /,
    *ai: _HasDType[_InexactMax32],
) -> type[np.complex64]: ...
@overload
def common_type(
    a0: _HasDType[np.float64],
    array1: _HasDType[_ComplexMax128],
    /,
    *ai: _HasDType[_NumberMax64],
) -> type[np.complex128]: ...
@overload
def common_type(
    a0: _HasDType[_ComplexMax128],
    array1: _HasDType[np.float64],
    /,
    *ai: _HasDType[_NumberMax64],
) -> type[np.complex128]: ...
@overload
def common_type(
    a0: _HasDType[_NumberMax64],
    array1: _HasDType[np.complex128],
    /,
    *ai: _HasDType[_NumberMax64],
) -> type[np.complex128]: ...
@overload
def common_type(
    a0: _HasDType[_ComplexMax128],
    array1: _HasDType[np.complex128 | np.integer],
    /,
    *ai: _HasDType[_NumberMax64],
) -> type[np.complex128]: ...
@overload
def common_type(
    a0: _HasDType[np.complex128 | np.integer],
    array1: _HasDType[_ComplexMax128],
    /,
    *ai: _HasDType[_NumberMax64],
) -> type[np.complex128]: ...
@overload
def common_type(
    a0: _HasDType[_Real],
    /,
    *ai: _HasDType[_Real],
) -> type[np.floating]: ...
@overload
def common_type(
    a0: _HasDType[np.number],
    array1: _HasDType[np.clongdouble],
    /,
    *ai: _HasDType[np.number],
) -> type[np.clongdouble]: ...
@overload
def common_type(
    a0: _HasDType[np.longdouble],
    array1: _HasDType[np.complexfloating],
    /,
    *ai: _HasDType[np.number],
) -> type[np.clongdouble]: ...
@overload
def common_type(
    a0: _HasDType[np.complexfloating],
    array1: _HasDType[np.longdouble],
    /,
    *ai: _HasDType[np.number],
) -> type[np.clongdouble]: ...
@overload
def common_type(
    a0: _HasDType[np.complexfloating],
    array1: _HasDType[np.number],
    /,
    *ai: _HasDType[np.number],
) -> type[np.complexfloating]: ...
@overload
def common_type(
    a0: _HasDType[np.number],
    array1: _HasDType[np.complexfloating],
    /,
    *ai: _HasDType[np.number],
) -> type[np.complexfloating]: ...
@overload
def common_type(
    a0: _HasDType[np.number],
    array1: _HasDType[np.number],
    /,
    *ai: _HasDType[np.number],
) -> type[Any]: ...



from __future__ import annotations

import io
from typing import TYPE_CHECKING, Any

from bokeh.io import export_png, export_svg, show
from bokeh.io.export import get_screenshot_as_png
from bokeh.layouts import gridplot
from bokeh.models.annotations.labels import Label
from bokeh.palettes import Category10
from bokeh.plotting import figure
import numpy as np

from contourpy.enum_util import as_fill_type, as_line_type
from contourpy.util.bokeh_util import filled_to_bokeh, lines_to_bokeh
from contourpy.util.renderer import Renderer

if TYPE_CHECKING:
    from bokeh.core.enums import OutputBackendType
    from bokeh.models import GridPlot
    from bokeh.palettes import Palette
    from numpy.typing import ArrayLike
    from selenium.webdriver.remote.webdriver import WebDriver

    from contourpy import FillType, LineType
    from contourpy._contourpy import FillReturn, LineReturn


class BokehRenderer(Renderer):
    """Utility renderer using Bokeh to render a grid of plots over the same (x, y) range.

    Args:
        nrows (int, optional): Number of rows of plots, default ``1``.
        ncols (int, optional): Number of columns of plots, default ``1``.
        figsize (tuple(float, float), optional): Figure size in inches (assuming 100 dpi), default
            ``(9, 9)``.
        show_frame (bool, optional): Whether to show frame and axes ticks, default ``True``.
        want_svg (bool, optional): Whether output is required in SVG format or not, default
            ``False``.

    Warning:
        :class:`~.BokehRenderer`, unlike :class:`~.MplRenderer`, needs to be told in advance if
        output to SVG format will be required later, otherwise it will assume PNG output.
    """
    _figures: list[figure]
    _layout: GridPlot
    _palette: Palette
    _want_svg: bool

    def __init__(
        self,
        nrows: int = 1,
        ncols: int = 1,
        figsize: tuple[float, float] = (9, 9),
        show_frame: bool = True,
        want_svg: bool = False,
    ) -> None:
        self._want_svg = want_svg
        self._palette = Category10[10]

        total_size = 100*np.asarray(figsize, dtype=int)  # Assuming 100 dpi.

        nfigures = nrows*ncols
        self._figures = []
        backend: OutputBackendType = "svg" if self._want_svg else "canvas"
        for _ in range(nfigures):
            fig = figure(output_backend=backend)
            fig.xgrid.visible = False
            fig.ygrid.visible = False
            self._figures.append(fig)
            if not show_frame:
                fig.outline_line_color = None
                fig.axis.visible = False

        self._layout = gridplot(
            self._figures, ncols=ncols, toolbar_location=None,  # type: ignore[arg-type]
            width=total_size[0] // ncols, height=total_size[1] // nrows)

    def _convert_color(self, color: str) -> str:
        if isinstance(color, str) and color[0] == "C":
            index = int(color[1:])
            color = self._palette[index]
        return color

    def _get_figure(self, ax: figure | int) -> figure:
        if isinstance(ax, int):
            ax = self._figures[ax]
        return ax

    def filled(
        self,
        filled: FillReturn,
        fill_type: FillType | str,
        ax: figure | int = 0,
        color: str = "C0",
        alpha: float = 0.7,
    ) -> None:
        """Plot filled contours on a single plot.

        Args:
            filled (sequence of arrays): Filled contour data as returned by
                :meth:`~.ContourGenerator.filled`.
            fill_type (FillType or str): Type of :meth:`~.ContourGenerator.filled` data as returned
                by :attr:`~.ContourGenerator.fill_type`, or a string equivalent.
            ax (int or Bokeh Figure, optional): Which plot to use, default ``0``.
            color (str, optional): Color to plot with. May be a string color or the letter ``"C"``
                followed by an integer in the range ``"C0"`` to ``"C9"`` to use a color from the
                ``Category10`` palette. Default ``"C0"``.
            alpha (float, optional): Opacity to plot with, default ``0.7``.
        """
        fill_type = as_fill_type(fill_type)
        fig = self._get_figure(ax)
        color = self._convert_color(color)
        xs, ys = filled_to_bokeh(filled, fill_type)
        if len(xs) > 0:
            fig.multi_polygons(xs=[xs], ys=[ys], color=color, fill_alpha=alpha, line_width=0)  # type: ignore[arg-type]

    def grid(
        self,
        x: ArrayLike,
        y: ArrayLike,
        ax: figure | int = 0,
        color: str = "black",
        alpha: float = 0.1,
        point_color: str | None = None,
        quad_as_tri_alpha: float = 0,
    ) -> None:
        """Plot quad grid lines on a single plot.

        Args:
            x (array-like of shape (ny, nx) or (nx,)): The x-coordinates of the grid points.
            y (array-like of shape (ny, nx) or (ny,)): The y-coordinates of the grid points.
            ax (int or Bokeh Figure, optional): Which plot to use, default ``0``.
            color (str, optional): Color to plot grid lines, default ``"black"``.
            alpha (float, optional): Opacity to plot lines with, default ``0.1``.
            point_color (str, optional): Color to plot grid points or ``None`` if grid points
                should not be plotted, default ``None``.
            quad_as_tri_alpha (float, optional): Opacity to plot ``quad_as_tri`` grid, default
                ``0``.

        Colors may be a string color or the letter ``"C"`` followed by an integer in the range
        ``"C0"`` to ``"C9"`` to use a color from the ``Category10`` palette.

        Warning:
            ``quad_as_tri_alpha > 0`` plots all quads as though they are unmasked.
        """
        fig = self._get_figure(ax)
        x, y = self._grid_as_2d(x, y)
        xs = list(x) + list(x.T)
        ys = list(y) + list(y.T)
        kwargs = {"line_color": color, "alpha": alpha}
        fig.multi_line(xs, ys, **kwargs)
        if quad_as_tri_alpha > 0:
            # Assumes no quad mask.
            xmid = (0.25*(x[:-1, :-1] + x[1:, :-1] + x[:-1, 1:] + x[1:, 1:])).ravel()
            ymid = (0.25*(y[:-1, :-1] + y[1:, :-1] + y[:-1, 1:] + y[1:, 1:])).ravel()
            fig.multi_line(
                list(np.stack((x[:-1, :-1].ravel(), xmid, x[1:, 1:].ravel()), axis=1)),
                list(np.stack((y[:-1, :-1].ravel(), ymid, y[1:, 1:].ravel()), axis=1)),
                **kwargs)
            fig.multi_line(
                list(np.stack((x[:-1, 1:].ravel(), xmid, x[1:, :-1].ravel()), axis=1)),
                list(np.stack((y[:-1, 1:].ravel(), ymid, y[1:, :-1].ravel()), axis=1)),
                **kwargs)
        if point_color is not None:
            fig.scatter(
                x=x.ravel(), y=y.ravel(), fill_color=color, line_color=None, alpha=alpha,
                marker="circle", size=8)

    def lines(
        self,
        lines: LineReturn,
        line_type: LineType | str,
        ax: figure | int = 0,
        color: str = "C0",
        alpha: float = 1.0,
        linewidth: float = 1,
    ) -> None:
        """Plot contour lines on a single plot.

        Args:
            lines (sequence of arrays): Contour line data as returned by
                :meth:`~.ContourGenerator.lines`.
            line_type (LineType or str): Type of :meth:`~.ContourGenerator.lines` data as returned
                by :attr:`~.ContourGenerator.line_type`, or a string equivalent.
            ax (int or Bokeh Figure, optional): Which plot to use, default ``0``.
            color (str, optional): Color to plot lines. May be a string color or the letter ``"C"``
                followed by an integer in the range ``"C0"`` to ``"C9"`` to use a color from the
                ``Category10`` palette. Default ``"C0"``.
            alpha (float, optional): Opacity to plot lines with, default ``1.0``.
            linewidth (float, optional): Width of lines, default ``1``.

        Note:
            Assumes all lines are open line strips not closed line loops.
        """
        line_type = as_line_type(line_type)
        fig = self._get_figure(ax)
        color = self._convert_color(color)
        xs, ys = lines_to_bokeh(lines, line_type)
        if xs is not None:
            assert ys is not None
            fig.line(xs, ys, line_color=color, line_alpha=alpha, line_width=linewidth)

    def mask(
        self,
        x: ArrayLike,
        y: ArrayLike,
        z: ArrayLike | np.ma.MaskedArray[Any, Any],
        ax: figure | int = 0,
        color: str = "black",
    ) -> None:
        """Plot masked out grid points as circles on a single plot.

        Args:
            x (array-like of shape (ny, nx) or (nx,)): The x-coordinates of the grid points.
            y (array-like of shape (ny, nx) or (ny,)): The y-coordinates of the grid points.
            z (masked array of shape (ny, nx): z-values.
            ax (int or Bokeh Figure, optional): Which plot to use, default ``0``.
            color (str, optional): Circle color, default ``"black"``.
        """
        mask = np.ma.getmask(z)
        if mask is np.ma.nomask:
            return
        fig = self._get_figure(ax)
        color = self._convert_color(color)
        x, y = self._grid_as_2d(x, y)
        fig.scatter(x[mask], y[mask], fill_color=color, marker="circle", size=10)

    def save(
        self,
        filename: str,
        transparent: bool = False,
        *,
        webdriver: WebDriver | None = None,
    ) -> None:
        """Save plots to SVG or PNG file.

        Args:
            filename (str): Filename to save to.
            transparent (bool, optional): Whether background should be transparent, default
                ``False``.
            webdriver (WebDriver, optional): Selenium WebDriver instance to use to create the image.

                .. versionadded:: 1.1.1

        Warning:
            To output to SVG file, ``want_svg=True`` must have been passed to the constructor.
        """
        if transparent:
            for fig in self._figures:
                fig.background_fill_color = None
                fig.border_fill_color = None

        if self._want_svg:
            export_svg(self._layout, filename=filename, webdriver=webdriver)
        else:
            export_png(self._layout, filename=filename, webdriver=webdriver)

    def save_to_buffer(self, *, webdriver: WebDriver | None = None) -> io.BytesIO:
        """Save plots to an ``io.BytesIO`` buffer.

        Args:
            webdriver (WebDriver, optional): Selenium WebDriver instance to use to create the image.

                .. versionadded:: 1.1.1

        Return:
            BytesIO: PNG image buffer.
        """
        image = get_screenshot_as_png(self._layout, driver=webdriver)
        buffer = io.BytesIO()
        image.save(buffer, "png")
        return buffer

    def show(self) -> None:
        """Show plots in web browser, in usual Bokeh manner.
        """
        show(self._layout)

    def title(self, title: str, ax: figure | int = 0, color: str | None = None) -> None:
        """Set the title of a single plot.

        Args:
            title (str): Title text.
            ax (int or Bokeh Figure, optional): Which plot to set the title of, default ``0``.
            color (str, optional): Color to set title. May be a string color or the letter ``"C"``
                followed by an integer in the range ``"C0"`` to ``"C9"`` to use a color from the
                ``Category10`` palette. Default ``None`` which is ``black``.
        """
        fig = self._get_figure(ax)
        fig.title = title
        fig.title.align = "center"  # type: ignore[attr-defined]
        if color is not None:
            fig.title.text_color = self._convert_color(color)  # type: ignore[attr-defined]

    def z_values(
        self,
        x: ArrayLike,
        y: ArrayLike,
        z: ArrayLike,
        ax: figure | int = 0,
        color: str = "green",
        fmt: str = ".1f",
        quad_as_tri: bool = False,
    ) -> None:
        """Show ``z`` values on a single plot.

        Args:
            x (array-like of shape (ny, nx) or (nx,)): The x-coordinates of the grid points.
            y (array-like of shape (ny, nx) or (ny,)): The y-coordinates of the grid points.
            z (array-like of shape (ny, nx): z-values.
            ax (int or Bokeh Figure, optional): Which plot to use, default ``0``.
            color (str, optional): Color of added text. May be a string color or the letter ``"C"``
                followed by an integer in the range ``"C0"`` to ``"C9"`` to use a color from the
                ``Category10`` palette. Default ``"green"``.
            fmt (str, optional): Format to display z-values, default ``".1f"``.
            quad_as_tri (bool, optional): Whether to show z-values at the ``quad_as_tri`` centres
                of quads.

        Warning:
            ``quad_as_tri=True`` shows z-values for all quads, even if masked.
        """
        fig = self._get_figure(ax)
        color = self._convert_color(color)
        x, y = self._grid_as_2d(x, y)
        z = np.asarray(z)
        ny, nx = z.shape
        kwargs = {"text_color": color, "text_align": "center", "text_baseline": "middle"}
        for j in range(ny):
            for i in range(nx):
                label = Label(x=x[j, i], y=y[j, i], text=f"{z[j, i]:{fmt}}", **kwargs)  # type: ignore[arg-type]
                fig.add_layout(label)
        if quad_as_tri:
            for j in range(ny-1):
                for i in range(nx-1):
                    xx = np.mean(x[j:j+2, i:i+2])
                    yy = np.mean(y[j:j+2, i:i+2])
                    zz = np.mean(z[j:j+2, i:i+2])
                    fig.add_layout(Label(x=xx, y=yy, text=f"{zz:{fmt}}", **kwargs))  # type: ignore[arg-type]



import os
import argparse
import logging
import shutil
import multiprocessing as mp
from contextlib import closing
from functools import partial

import fontTools
from .ufo import font_to_quadratic, fonts_to_quadratic

ufo_module = None
try:
    import ufoLib2 as ufo_module
except ImportError:
    try:
        import defcon as ufo_module
    except ImportError as e:
        pass


logger = logging.getLogger("fontTools.cu2qu")


def _cpu_count():
    try:
        return mp.cpu_count()
    except NotImplementedError:  # pragma: no cover
        return 1


def open_ufo(path):
    if hasattr(ufo_module.Font, "open"):  # ufoLib2
        return ufo_module.Font.open(path)
    return ufo_module.Font(path)  # defcon


def _font_to_quadratic(input_path, output_path=None, **kwargs):
    ufo = open_ufo(input_path)
    logger.info("Converting curves for %s", input_path)
    if font_to_quadratic(ufo, **kwargs):
        logger.info("Saving %s", output_path)
        if output_path:
            ufo.save(output_path)
        else:
            ufo.save()  # save in-place
    elif output_path:
        _copytree(input_path, output_path)


def _samepath(path1, path2):
    # TODO on python3+, there's os.path.samefile
    path1 = os.path.normcase(os.path.abspath(os.path.realpath(path1)))
    path2 = os.path.normcase(os.path.abspath(os.path.realpath(path2)))
    return path1 == path2


def _copytree(input_path, output_path):
    if _samepath(input_path, output_path):
        logger.debug("input and output paths are the same file; skipped copy")
        return
    if os.path.exists(output_path):
        shutil.rmtree(output_path)
    shutil.copytree(input_path, output_path)


def _main(args=None):
    """Convert a UFO font from cubic to quadratic curves"""
    parser = argparse.ArgumentParser(prog="cu2qu")
    parser.add_argument("--version", action="version", version=fontTools.__version__)
    parser.add_argument(
        "infiles",
        nargs="+",
        metavar="INPUT",
        help="one or more input UFO source file(s).",
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument(
        "-e",
        "--conversion-error",
        type=float,
        metavar="ERROR",
        default=None,
        help="maximum approximation error measured in EM (default: 0.001)",
    )
    parser.add_argument(
        "-m",
        "--mixed",
        default=False,
        action="store_true",
        help="whether to used mixed quadratic and cubic curves",
    )
    parser.add_argument(
        "--keep-direction",
        dest="reverse_direction",
        action="store_false",
        help="do not reverse the contour direction",
    )

    mode_parser = parser.add_mutually_exclusive_group()
    mode_parser.add_argument(
        "-i",
        "--interpolatable",
        action="store_true",
        help="whether curve conversion should keep interpolation compatibility",
    )
    mode_parser.add_argument(
        "-j",
        "--jobs",
        type=int,
        nargs="?",
        default=1,
        const=_cpu_count(),
        metavar="N",
        help="convert using N multiple processes (default: %(default)s)",
    )

    output_parser = parser.add_mutually_exclusive_group()
    output_parser.add_argument(
        "-o",
        "--output-file",
        default=None,
        metavar="OUTPUT",
        help=(
            "output filename for the converted UFO. By default fonts are "
            "modified in place. This only works with a single input."
        ),
    )
    output_parser.add_argument(
        "-d",
        "--output-dir",
        default=None,
        metavar="DIRECTORY",
        help="output directory where to save converted UFOs",
    )

    options = parser.parse_args(args)

    if ufo_module is None:
        parser.error("Either ufoLib2 or defcon are required to run this script.")

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level)

    if len(options.infiles) > 1 and options.output_file:
        parser.error("-o/--output-file can't be used with multile inputs")

    if options.output_dir:
        output_dir = options.output_dir
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        elif not os.path.isdir(output_dir):
            parser.error("'%s' is not a directory" % output_dir)
        output_paths = [
            os.path.join(output_dir, os.path.basename(p)) for p in options.infiles
        ]
    elif options.output_file:
        output_paths = [options.output_file]
    else:
        # save in-place
        output_paths = [None] * len(options.infiles)

    kwargs = dict(
        dump_stats=options.verbose > 0,
        max_err_em=options.conversion_error,
        reverse_direction=options.reverse_direction,
        all_quadratic=False if options.mixed else True,
    )

    if options.interpolatable:
        logger.info("Converting curves compatibly")
        ufos = [open_ufo(infile) for infile in options.infiles]
        if fonts_to_quadratic(ufos, **kwargs):
            for ufo, output_path in zip(ufos, output_paths):
                logger.info("Saving %s", output_path)
                if output_path:
                    ufo.save(output_path)
                else:
                    ufo.save()
        else:
            for input_path, output_path in zip(options.infiles, output_paths):
                if output_path:
                    _copytree(input_path, output_path)
    else:
        jobs = min(len(options.infiles), options.jobs) if options.jobs > 1 else 1
        if jobs > 1:
            func = partial(_font_to_quadratic, **kwargs)
            logger.info("Running %d parallel processes", jobs)
            with closing(mp.Pool(jobs)) as pool:
                pool.starmap(func, zip(options.infiles, output_paths))
        else:
            for input_path, output_path in zip(options.infiles, output_paths):
                _font_to_quadratic(input_path, output_path, **kwargs)



"""Tests for the two-sided pooled two-proportion z-test."""

from __future__ import annotations

import math

from scipy.stats import norm

from abtestwise.frequentist import two_proportion_z_test


def test_known_case_matches_manual_calculation():
    # control 120/1000, treatment 145/1000
    z, p = two_proportion_z_test(120, 1000, 145, 1000)

    pooled = (120 + 145) / 2000
    se = math.sqrt(pooled * (1 - pooled) * (1 / 1000 + 1 / 1000))
    expected_z = (0.145 - 0.120) / se
    expected_p = 2 * (1 - norm.cdf(abs(expected_z)))

    assert abs(z - expected_z) < 1e-9
    assert abs(p - expected_p) < 1e-12


def test_treatment_higher_gives_positive_z():
    z, _ = two_proportion_z_test(100, 1000, 150, 1000)
    assert z > 0


def test_swapping_arms_flips_sign_keeps_pvalue():
    z1, p1 = two_proportion_z_test(120, 1000, 145, 1000)
    z2, p2 = two_proportion_z_test(145, 1000, 120, 1000)
    assert abs(z1 + z2) < 1e-12
    assert abs(p1 - p2) < 1e-12


def test_identical_rates_give_zero_z_and_p_one():
    z, p = two_proportion_z_test(100, 1000, 100, 1000)
    assert abs(z) < 1e-12
    assert abs(p - 1.0) < 1e-12


def test_degenerate_zero_standard_error():
    # Both arms all failures -> pooled rate 0 -> se 0 -> defined fallback.
    z, p = two_proportion_z_test(0, 1000, 0, 1000)
    assert z == 0.0
    assert p == 1.0



"""
Platform independent script to download all the
`scipy.datasets` module data files.
This doesn't require a full scipy build.

Run: python _download_all.py <download_dir>
"""

from scipy._lib._array_api import xp_capabilities

import argparse
try:
    import pooch
except ImportError:
    pooch = None


if __spec__.parent is None or __spec__.parent == '':
    # Running as python script, use absolute import
    import _registry  # type: ignore
else:
    # Running as python module, use relative import
    from . import _registry


@xp_capabilities(out_of_scope=True)
def download_all(path=None):
    """
    Utility method to download all the dataset files
    for `scipy.datasets` module.

    Parameters
    ----------
    path : str, optional
        Directory path to download all the dataset files.
        If None, default to the system cache_dir detected by pooch.

    Examples
    --------
    Download the datasets to the default cache location:

    >>> from scipy import datasets
    >>> datasets.download_all()

    Download the datasets to the current directory:

    >>> datasets.download_all(".")

    """
    if pooch is None:
        raise ImportError("Missing optional dependency 'pooch' required "
                          "for scipy.datasets module. Please use pip or "
                          "conda to install 'pooch'.")
    if path is None:
        path = pooch.os_cache('scipy-data')
    # https://github.com/scipy/scipy/issues/21879
    downloader = pooch.HTTPDownloader(headers={"User-Agent": "SciPy"})
    for dataset_name, dataset_hash in _registry.registry.items():
        pooch.retrieve(url=_registry.registry_urls[dataset_name],
                       known_hash=dataset_hash,
                       fname=dataset_name, path=path, downloader=downloader)


def main():
    parser = argparse.ArgumentParser(description='Download SciPy data files.')
    parser.add_argument("path", nargs='?', type=str,
                        default=pooch.os_cache('scipy-data'),
                        help="Directory path to download all the data files.")
    args = parser.parse_args()
    download_all(args.path)


if __name__ == "__main__":
    main()



"""
Tests related to deprecation warnings. Also a convenient place
to document how deprecations should eventually be turned into errors.

"""
import contextlib
import re
import warnings
from collections.abc import Callable

import pytest

import numpy as np
from numpy._core._multiarray_tests import fromstring_null_term_c_api  # noqa: F401
from numpy.testing import assert_raises


class _DeprecationTestCase:
    # Just as warning: warnings uses re.match, so the start of this message
    # must match.
    message = ''
    warning_cls = DeprecationWarning

    @contextlib.contextmanager
    def filter_warnings(self):
        with warnings.catch_warnings(record=True) as w:
            # Do *not* ignore other DeprecationWarnings. Ignoring warnings
            # can give very confusing results because of
            # https://bugs.python.org/issue4180 and it is probably simplest to
            # try to keep the tests cleanly giving only the right warning type.
            # (While checking them set to "error" those are ignored anyway)
            # We still have them show up, because otherwise they would be raised
            warnings.filterwarnings("always", category=self.warning_cls)
            warnings.filterwarnings("always", message=self.message,
                                    category=self.warning_cls)
            yield w
        return

    def assert_deprecated(self, function, num=1, ignore_others=False,
                          function_fails=False,
                          exceptions=np._NoValue,
                          args=(), kwargs={}):
        """Test if DeprecationWarnings are given and raised.

        This first checks if the function when called gives `num`
        DeprecationWarnings, after that it tries to raise these
        DeprecationWarnings and compares them with `exceptions`.
        The exceptions can be different for cases where this code path
        is simply not anticipated and the exception is replaced.

        Parameters
        ----------
        function : callable
            The function to test
        num : int
            Number of DeprecationWarnings to expect. This should normally be 1.
        ignore_others : bool
            Whether warnings of the wrong type should be ignored (note that
            the message is not checked)
        function_fails : bool
            If the function would normally fail, setting this will check for
            warnings inside a try/except block.
        exceptions : Exception or tuple of Exceptions
            Exception to expect when turning the warnings into an error.
            The default checks for DeprecationWarnings. If exceptions is
            empty the function is expected to run successfully.
        args : tuple
            Arguments for `function`
        kwargs : dict
            Keyword arguments for `function`
        """
        __tracebackhide__ = True  # Hide traceback for py.test

        if exceptions is np._NoValue:
            exceptions = (self.warning_cls,)

        if function_fails:
            context_manager = contextlib.suppress(Exception)
        else:
            context_manager = contextlib.nullcontext()
        with context_manager:
            with self.filter_warnings() as w_context:
                function(*args, **kwargs)

        # just in case, clear the registry
        num_found = 0
        for warning in w_context:
            if warning.category is self.warning_cls:
                num_found += 1
                if self.message:
                    if not re.match(self.message, str(warning.message)):
                        raise AssertionError(
                            f"Warning message '{warning.message}' did not match "
                            f"expected pattern '{self.message}'"
                        )
            elif not ignore_others:
                name = self.warning_cls.__name__
                raise AssertionError(f"expected {name} but got: {warning.category}")
        if num is not None and num_found != num:
            msg = f"{len(w_context)} warnings found but {num} expected."
            lst = [str(w) for w in w_context]
            raise AssertionError("\n".join([msg] + lst))

        with warnings.catch_warnings():
            warnings.filterwarnings("error", message=self.message,
                                    category=self.warning_cls)
            try:
                function(*args, **kwargs)
                if exceptions != ():
                    raise AssertionError(
                            "No error raised during function call")
            except exceptions:
                if exceptions == ():
                    raise AssertionError(
                            "Error raised during function call")

    def assert_not_deprecated(self, function, args=(), kwargs={}):
        """Test that warnings are not raised.

        This is just a shorthand for:

        self.assert_deprecated(function, num=0, ignore_others=True,
                        exceptions=tuple(), args=args, kwargs=kwargs)
        """
        self.assert_deprecated(function, num=0, ignore_others=True,
                        exceptions=(), args=args, kwargs=kwargs)


class _VisibleDeprecationTestCase(_DeprecationTestCase):
    warning_cls = np.exceptions.VisibleDeprecationWarning


class TestTestDeprecated:
    def test_assert_deprecated(self):
        test_case_instance = _DeprecationTestCase()
        assert_raises(AssertionError,
                      test_case_instance.assert_deprecated,
                      lambda: None)

        def foo():
            warnings.warn("foo", category=DeprecationWarning, stacklevel=2)

        test_case_instance.assert_deprecated(foo)


class TestCtypesGetter(_DeprecationTestCase):
    ctypes = np.array([1]).ctypes

    @pytest.mark.parametrize("name", ["data", "shape", "strides", "_as_parameter_"])
    def test_not_deprecated(self, name: str) -> None:
        self.assert_not_deprecated(lambda: getattr(self.ctypes, name))


class TestPyIntConversion(_DeprecationTestCase):
    message = r".*stop allowing conversion of out-of-bound.*"

    @pytest.mark.parametrize("dtype", np.typecodes["AllInteger"])
    def test_deprecated_scalar(self, dtype):
        dtype = np.dtype(dtype)
        info = np.iinfo(dtype)

        # Cover the most common creation paths (all end up in the
        # same place):
        def scalar(value, dtype):
            dtype.type(value)

        def assign(value, dtype):
            arr = np.array([0, 0, 0], dtype=dtype)
            arr[2] = value

        def create(value, dtype):
            np.array([value], dtype=dtype)

        for creation_func in [scalar, assign, create]:
            try:
                self.assert_deprecated(
                        lambda: creation_func(info.min - 1, dtype))
            except OverflowError:
                pass  # OverflowErrors always happened also before and are OK.

            try:
                self.assert_deprecated(
                        lambda: creation_func(info.max + 1, dtype))
            except OverflowError:
                pass  # OverflowErrors always happened also before and are OK.


@pytest.mark.parametrize("name", ["str", "bytes", "object"])
def test_future_scalar_attributes(name):
    # FutureWarning added 2022-11-17, NumPy 1.24,
    assert name not in dir(np)  # we may want to not add them
    with pytest.warns(FutureWarning,
            match=f"In the future .*{name}"):
        assert not hasattr(np, name)

    # Unfortunately, they are currently still valid via `np.dtype()`
    np.dtype(name)
    name in np._core.sctypeDict


# Ignore the above future attribute warning for this test.
@pytest.mark.filterwarnings("ignore:In the future:FutureWarning")
class TestRemovedGlobals:
    # Removed 2023-01-12, NumPy 1.24.0
    # Not a deprecation, but the large error was added to aid those who missed
    # the previous deprecation, and should be removed similarly to one
    # (or faster).
    @pytest.mark.parametrize("name",
            ["object", "float", "complex", "str", "int"])
    def test_attributeerror_includes_info(self, name):
        msg = f".*\n`np.{name}` was a deprecated alias for the builtin"
        with pytest.raises(AttributeError, match=msg):
            getattr(np, name)


class TestCharArray(_DeprecationTestCase):
    def test_deprecated_chararray(self):
        self.assert_deprecated(lambda: np.char.chararray)


class TestDeprecatedDTypeAliases(_DeprecationTestCase):
    @pytest.mark.parametrize("dtype_code", ["a", "a10"])
    def test_a_dtype_alias(self, dtype_code: str):
        # Deprecated in 2.0, removed in 2.5, 2025-12
        with pytest.raises(TypeError):
            np.dtype(dtype_code)


class TestDeprecatedArrayWrap(_DeprecationTestCase):
    message = "__array_wrap__.*"

    def test_deprecated(self):
        class Test1:
            def __array__(self, dtype=None, copy=None):
                return np.arange(4)

            def __array_wrap__(self, arr, context=None):
                self.called = True
                return 'pass context'

        class Test2(Test1):
            def __array_wrap__(self, arr):
                self.called = True
                return 'pass'

        test1 = Test1()
        test2 = Test2()
        self.assert_deprecated(lambda: np.negative(test1))
        assert test1.called
        self.assert_deprecated(lambda: np.negative(test2))
        assert test2.called


class TestDeprecatedArrayAttributeSetting(_DeprecationTestCase):
    message = "Setting the .*on a NumPy array has been deprecated.*"

    def test_deprecated_strides_set(self):
        x = np.eye(2)
        self.assert_deprecated(setattr, args=(x, 'strides', x.strides))

    def test_deprecated_dtype_set(self):
        x = np.eye(2)
        self.assert_deprecated(setattr, args=(x, "dtype", int))

    def test_deprecated_dtype_set_record(self):
        x = np.zeros(2, dtype="i4,i4").view(np.recarray)
        self.assert_deprecated(setattr, args=(x, "dtype", np.dtype("f4,f4")))

    def test_deprecated_shape_set(self):
        x = np.eye(2)
        self.assert_deprecated(setattr, args=(x, "shape", (4, 1)))


class TestDeprecatedViewDtypePropertySetter(_DeprecationTestCase):
    # view() with dtype change on a subclass that overrides the
    # dtype property should warn to implement _set_dtype instead.
    message = r"numpy.ndarray.view\(\) used a custom `dtype` setter.*"

    def test_view_dtype_property_setter(self):
        class MyArray(np.ndarray):
            @property
            def dtype(self):
                return super().dtype

            @dtype.setter
            def dtype(self, dtype):
                super(MyArray, type(self))._set_dtype(self, dtype)

        arr = np.arange(6).view(MyArray)
        self.assert_deprecated(arr.view, args=(np.float64,))

    def test_view_dtype_property_setter_recarray_subclass(self):
        # Recarray subclasses without a `_set_dtype` should still work
        # but warn until they implement it.  As recarray sets it to None
        # such a subclass may have to do the same fix here (can't use super()).
        class SideBranch:
            pass

        class MyRecArray(SideBranch, np.recarray):
            @property
            def dtype(self):
                return super().dtype

            @dtype.setter
            def dtype(self, dtype):
                # would need to call `np.ndarray._set_dtype` to avoid the warning
                # (but that side-steps the recarray dtype handling.)
                with pytest.warns(DeprecationWarning, match="Setting the dtype"):
                    np.recarray.dtype.__set__(self, dtype)

        arr = np.rec.fromarrays(
                [np.arange(3, dtype=np.int32)], names='x').view(MyRecArray)
        self.assert_deprecated(arr.view, args=([('y', 'i4')],))


class TestDeprecatedDTypeParenthesizedRepeatCount(_DeprecationTestCase):
    message = "Passing in a parenthesized single number"

    @pytest.mark.parametrize("string", ["(2)i,", "(3)3S,", "f,(2)f"])
    def test_parenthesized_repeat_count(self, string):
        self.assert_deprecated(np.dtype, args=(string,))


class TestDTypeAlignBool(_VisibleDeprecationTestCase):
    # Deprecated in Numpy 2.4, 2025-07
    # NOTE: As you can see, finalizing this deprecation breaks some (very) old
    # pickle files.  This may be fine, but needs to be done with some care since
    # it breaks all of them and not just some.
    # (Maybe it should be a 3.0 or only after warning more explicitly around pickles.)
    message = r"dtype\(\): align should be passed as Python or NumPy boolean but got "

    def test_deprecated(self):
        # in particular integers should be rejected because one may think they mean
        # alignment, or pass them accidentally as a subarray shape (meaning to pass
        # a tuple).
        self.assert_deprecated(lambda: np.dtype("f8", align=3))
        self.assert_deprecated(lambda: np.dtype("f8", align=0, copy=10**100))
        self.assert_deprecated(lambda: np.dtype("f8", align=10**100, copy=0))
        # Subclasses of ints don't hit the below pickle code path:
        self.assert_deprecated(
            lambda: np.dtype("f8", align=np.long(0), copy=np.long(1)))

    @pytest.mark.parametrize("align", [True, False, np.True_, np.False_])
    def test_not_deprecated(self, align):
        # if the user passes a bool, it is accepted.
        self.assert_not_deprecated(lambda: np.dtype("f8", align=align))
        # The following specific case is used by old pickles:
        self.assert_not_deprecated(lambda: np.dtype("f8", align=0, copy=1))


class TestFlatiterIndexing0dBoolIndex(_DeprecationTestCase):
    # Deprecated in Numpy 2.4, 2025-07
    message = r"Indexing flat iterators with a 0-dimensional boolean index"

    def test_0d_boolean_index_deprecated(self):
        arr = np.arange(3)
        # 0d boolean indices on flat iterators are deprecated
        self.assert_deprecated(lambda: arr.flat[True])

    def test_0d_boolean_assign_index_deprecated(self):
        arr = np.arange(3)

        def assign_to_index():
            arr.flat[True] = 10

        self.assert_deprecated(assign_to_index)


class TestFlatiterIndexingFloatIndex(_DeprecationTestCase):
    # Deprecated in NumPy 2.4, 2025-07
    message = r"Invalid non-array indices for iterator objects"

    def test_float_index_deprecated(self):
        arr = np.arange(3)
        # float indices on flat iterators are deprecated
        self.assert_deprecated(lambda: arr.flat[[1.]])

    def test_float_assign_index_deprecated(self):
        arr = np.arange(3)

        def assign_to_index():
            arr.flat[[1.]] = 10

        self.assert_deprecated(assign_to_index)


@pytest.mark.thread_unsafe(
    reason="warning control utilities are deprecated due to being thread-unsafe"
)
class TestWarningUtilityDeprecations(_DeprecationTestCase):
    # Deprecation in NumPy 2.4, 2025-08
    message = r"NumPy warning suppression and assertion utilities are deprecated."

    def test_assert_warns_deprecated(self):
        def use_assert_warns():
            with np.testing.assert_warns(RuntimeWarning):
                warnings.warn("foo", RuntimeWarning, stacklevel=1)

        self.assert_deprecated(use_assert_warns)

    def test_suppress_warnings_deprecated(self):
        def use_suppress_warnings():
            with np.testing.suppress_warnings() as sup:
                sup.filter(RuntimeWarning, 'invalid value encountered in divide')

        self.assert_deprecated(use_suppress_warnings)


class TestTooManyArgsExtremum(_DeprecationTestCase):
    # Deprecated in Numpy 2.4, 2025-08, gh-27639
    message = "Passing more than 2 positional arguments to np.maximum and np.minimum "

    @pytest.mark.parametrize("ufunc", [np.minimum, np.maximum])
    def test_extremum_3_args(self, ufunc):
        self.assert_deprecated(ufunc, args=(np.ones(1), np.zeros(1), np.empty(1)))


class TestTypenameDeprecation(_DeprecationTestCase):
    # Deprecation in Numpy 2.5, 2026-02

    def test_typename_emits_deprecation_warning(self):
        self.assert_deprecated(lambda: np.typename("S1"))
        self.assert_deprecated(lambda: np.typename("h"))

class TestRoundDeprecation(_DeprecationTestCase):
    # Deprecation in NumPy 2.5, 2026-02

    def test_round_emits_deprecation_warning_array(self):
        a = np.array([1.5, 2.7, -1.5, -2.7])
        self.assert_deprecated(lambda: np.ma.round_(a))

    def test_round_emits_deprecation_warning_scalar(self):
        self.assert_deprecated(lambda: np.ma.round_(3.14))


class TestDeprecatedGenericTimedelta(_DeprecationTestCase):
    # Deprecated in Numpy 2.5, 2025-11
    # See gh-29619
    message = "The 'generic' unit for NumPy timedelta is deprecated"

    @pytest.mark.parametrize('value', [
        3, 10, "NaT"
    ])
    def test_raise_warning_for_timedelta_with_generic_unit(self, value: int | str):
        self.assert_deprecated(lambda x: np.timedelta64(x), args=(value,))

    @pytest.mark.parametrize('value', [
        np.timedelta64(3, "s"), np.timedelta64(10, "D")
    ])
    @pytest.mark.parametrize('generic_value', [
        5, 2
    ])
    @pytest.mark.parametrize(
        "op",
        [
            np.add,
            np.subtract,
        ],
    )
    def test_raise_warning_for_operation_with_generic_unit(
        self, value: int, generic_value: int, op: Callable
    ):
        self.assert_deprecated(op, args=(value, generic_value))

    def test_raise_warning_for_default_constructor(self):
        self.assert_deprecated(lambda: np.timedelta64())
        self.assert_deprecated(lambda: np.datetime64())

    def test_raise_warning_for_NAT_construction(self):
        self.assert_deprecated(lambda: np.datetime64('NaT'))
        self.assert_deprecated(lambda: np.datetime64(None))


class TestTriDeprecationWithNonInteger(_DeprecationTestCase):
    # Deprecation in NumPy 2.5, 2026-03

    def test_tri(self):
        self.assert_deprecated(lambda: np.tri(M=2.3, k=3.14, N=np.object_(8)))

    def test_triu_indices(self):
        self.assert_deprecated(lambda: np.triu_indices(n=np.float64(7.14), k=3.2))
        self.assert_deprecated(lambda: np.triu_indices(n=4, k=np.bool(0)))

    def test_tril_indices(self):
        self.assert_deprecated(lambda: np.tril_indices(n=np.array(3.14)))

    def test_triu_indices_from(self):
        a = np.array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11],
           [12, 13, 14, 15]])
        self.assert_deprecated(lambda: np.triu_indices_from(a, k=np.object_(9.8)))

    def test_tril_indices_from(self):
        a = np.array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11],
           [12, 13, 14, 15]])
        self.assert_deprecated(lambda: np.tril_indices_from(a, k=9.8))


class TestTakeOutDtype(_DeprecationTestCase):
    # Deprecated in Numpy 2.5, 2026-01
    message = "Implicit casting of output to a different kind."

    def test_out_dtype_deprecated(self):
        a = np.arange(3).astype(np.int32)
        indices = np.arange(2)
        different_dtype_out = np.zeros_like(indices, dtype=np.uint32)

        self.assert_deprecated(lambda: np.take(a, indices, out=different_dtype_out))



"""xpkit: a lightweight toolkit for binary A/B experiment analysis.

Version 0.1 combines frequentist and Bayesian summaries for binary proportions
using aggregate count data.
"""

from __future__ import annotations

from .binary import BinaryABTest
from .result import BinaryABResult

__version__ = "0.1.0"

__all__ = ["BinaryABTest", "BinaryABResult", "__version__"]



import matplotlib.cm as cm
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.collections import Collection
from matplotlib.colorizer import Colorizer, ColorizingArtist
from matplotlib.colors import Colormap, Normalize
from matplotlib.path import Path
from matplotlib.patches import Patch
from matplotlib.text import Text
from matplotlib.transforms import Transform, TransformedPatchPath, TransformedPath
from matplotlib.ticker import Locator, Formatter

from numpy.typing import ArrayLike
import numpy as np
from collections.abc import Callable, Iterable, Sequence
from typing import Literal
from .typing import ColorType



class ContourLabeler:
    labelFmt: str | Formatter | Callable[[float], str] | dict[float, str]
    labelManual: bool | Iterable[tuple[float, float]]
    rightside_up: bool
    labelLevelList: list[float]
    labelIndiceList: list[int]
    labelMappable: cm.ScalarMappable | ColorizingArtist
    labelCValueList: list[ColorType]
    labelXYs: list[tuple[float, float]]
    def clabel(
        self,
        levels: ArrayLike | None = ...,
        *,
        fontsize: str | float | None = ...,
        inline: bool = ...,
        inline_spacing: float = ...,
        fmt: str | Formatter | Callable[[float], str] | dict[float, str] | None = ...,
        colors: ColorType | Sequence[ColorType] | None = ...,
        use_clabeltext: bool = ...,
        manual: bool | Iterable[tuple[float, float]] = ...,
        rightside_up: bool = ...,
        zorder: float | None = ...
    ) -> list[Text]: ...
    def print_label(self, linecontour: ArrayLike, labelwidth: float) -> bool: ...
    def too_close(self, x: float, y: float, lw: float) -> bool: ...
    def get_text(
        self,
        lev: float,
        fmt: str | Formatter | Callable[[float], str] | dict[float, str],
    ) -> str: ...
    def locate_label(
        self, linecontour: ArrayLike, labelwidth: float
    ) -> tuple[float, float, float]: ...
    def add_label(
        self, x: float, y: float, rotation: float, lev: float, cvalue: ColorType
    ) -> None: ...
    def add_label_near(
        self,
        x: float,
        y: float,
        inline: bool = ...,
        inline_spacing: int = ...,
        transform: Transform | Literal[False] | None = ...,
    ) -> None: ...
    def pop_label(self, index: int = ...) -> None: ...
    def labels(self, inline: bool, inline_spacing: int) -> None: ...
    def remove(self) -> None: ...

class ContourSet(ContourLabeler, Collection):
    axes: Axes
    levels: Iterable[float]
    filled: bool
    linewidths: float | ArrayLike | None
    hatches: Iterable[str | None]
    origin: Literal["upper", "lower", "image"] | None
    extent: tuple[float, float, float, float] | None
    colors: ColorType | Sequence[ColorType]
    extend: Literal["neither", "both", "min", "max"]
    nchunk: int
    locator: Locator | None
    logscale: bool
    negative_linestyles: None | Literal[
        "solid", "dashed", "dashdot", "dotted"
    ] | Iterable[Literal["solid", "dashed", "dashdot", "dotted"]]
    clip_path: Patch | Path | TransformedPath | TransformedPatchPath | None
    labelTexts: list[Text]
    labelCValues: list[ColorType]

    @property
    def allkinds(self) -> list[list[np.ndarray | None]]: ...
    @property
    def allsegs(self) -> list[list[np.ndarray]]: ...
    @property
    def alpha(self) -> float | None: ...
    @property
    def linestyles(self) -> (
        None |
        Literal["solid", "dashed", "dashdot", "dotted"] |
        Iterable[Literal["solid", "dashed", "dashdot", "dotted"]]
    ): ...

    def __init__(
        self,
        ax: Axes,
        *args,
        levels: Iterable[float] | None = ...,
        filled: bool = ...,
        linewidths: float | ArrayLike | None = ...,
        linestyles: Literal["solid", "dashed", "dashdot", "dotted"]
        | Iterable[Literal["solid", "dashed", "dashdot", "dotted"]]
        | None = ...,
        hatches: Iterable[str | None] = ...,
        alpha: float | None = ...,
        origin: Literal["upper", "lower", "image"] | None = ...,
        extent: tuple[float, float, float, float] | None = ...,
        cmap: str | Colormap | None = ...,
        colors: ColorType | Sequence[ColorType] | None = ...,
        norm: str | Normalize | None = ...,
        vmin: float | None = ...,
        vmax: float | None = ...,
        colorizer: Colorizer | None = ...,
        extend: Literal["neither", "both", "min", "max"] = ...,
        antialiased: bool | None = ...,
        nchunk: int = ...,
        locator: Locator | None = ...,
        transform: Transform | None = ...,
        negative_linestyles: Literal["solid", "dashed", "dashdot", "dotted"]
        | Iterable[Literal["solid", "dashed", "dashdot", "dotted"]]
        | None = ...,
        clip_path: Patch | Path | TransformedPath | TransformedPatchPath | None = ...,
        **kwargs
    ) -> None: ...
    def legend_elements(
        self, variable_name: str = ..., str_format: Callable[[float], str] = ...
    ) -> tuple[list[Artist], list[str]]: ...
    def find_nearest_contour(
        self, x: float, y: float, indices: Iterable[int] | None = ..., pixel: bool = ...
    ) -> tuple[int, int, int, float, float, float]: ...

class QuadContourSet(ContourSet): ...
